"""Test the translations API."""

from copy import deepcopy
from pathlib import Path

import pytest

from napari.utils.translations import (
    TranslationString,
    _get_display_name,
    _is_valid_locale,
    _load_language,
    get_language_packs,
    translator,
)

TEST_LOCALE = 'es_CO'
HERE = Path(__file__).parent
TEST_LANGUAGE_PACK_PATH = HERE / 'napari-language-pack-es_CO'


es_CO_po = r"""msgid ""
msgstr ""
"Project-Id-Version: \n"
"POT-Creation-Date: 2021-02-18 19:00\n"
"PO-Revision-Date: 2021-04-08 08:14-0500\n"
"Language-Team: \n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"X-Generator: Poedit 2.4.2\n"
"Last-Translator: \n"
"Plural-Forms: nplurals=2; plural=(n != 1);\n"
"Language: es_CO\n"

#: Test for singular
msgid "More about napari"
msgstr "Más sobre napari"

#: Test for singular with context
msgctxt "singular-context"
msgid "More about napari with context"
msgstr "Más sobre napari con contexto"

#: Test for singular with variables
msgid "More about napari with {variable}"
msgstr "Más sobre napari con {variable}"

#: Test for singular with and context variables
msgctxt "singular-context-variables"
msgid "More about napari with context and {variable}"
msgstr "Más sobre napari con contexto y {variable}"

#: Test for plural
msgid "I have napari"
msgid_plural "I have naparis"
msgstr[0] "Tengo napari"
msgstr[1] "Tengo naparis"

#: Test for singular with context
msgctxt "plural-context"
msgid "I have napari with context"
msgid_plural "I have naparis with context"
msgstr[0] "Tengo napari con contexto"
msgstr[1] "Tengo naparis con contexto"

#: Test for plural with variables
msgid "I have {n} napari with {variable}"
msgid_plural "I have {n} naparis with {variable}"
msgstr[0] "Tengo {n} napari con {variable}"
msgstr[1] "Tengo {n} naparis con {variable}"

#: Test for singular with and context variables
msgctxt "plural-context-variables"
msgid "I have {n} napari with {variable} and context"
msgid_plural "I have {n} naparis with {variable} and context"
msgstr[0] "Tengo {n} napari con {variable} y contexto"
msgstr[1] "Tengo {n} naparis con {variable} y contexto"
"""

es_CO_mo = b'\xde\x12\x04\x95\x00\x00\x00\x00\t\x00\x00\x00\x1c\x00\x00\x00d\x00\x00\x00\r\x00\x00\x00\xac\x00\x00\x00\x00\x00\x00\x00\xe0\x00\x00\x00\x1c\x00\x00\x00\xe1\x00\x00\x00D\x00\x00\x00\xfe\x00\x00\x00\x11\x00\x00\x00C\x01\x00\x00!\x00\x00\x00U\x01\x00\x00E\x00\x00\x00w\x01\x00\x00u\x00\x00\x00\xbd\x01\x00\x00/\x00\x00\x003\x02\x00\x00H\x00\x00\x00c\x02\x00\x00\x0e\x01\x00\x00\xac\x02\x00\x00\x1a\x00\x00\x00\xbb\x03\x00\x00@\x00\x00\x00\xd6\x03\x00\x00\x11\x00\x00\x00\x17\x04\x00\x00 \x00\x00\x00)\x04\x00\x004\x00\x00\x00J\x04\x00\x00V\x00\x00\x00\x7f\x04\x00\x00\x1e\x00\x00\x00\xd6\x04\x00\x00+\x00\x00\x00\xf5\x04\x00\x00\x01\x00\x00\x00\t\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x08\x00\x00\x00\x06\x00\x00\x00\x05\x00\x00\x00\x00\x00\x00\x00\x07\x00\x00\x00\x02\x00\x00\x00\x00I have napari\x00I have naparis\x00I have {n} napari with {variable}\x00I have {n} naparis with {variable}\x00More about napari\x00More about napari with {variable}\x00plural-context\x04I have napari with context\x00I have naparis with context\x00plural-context-variables\x04I have {n} napari with {variable} and context\x00I have {n} naparis with {variable} and context\x00singular-context\x04More about napari with context\x00singular-context-variables\x04More about napari with context and {variable}\x00Project-Id-Version: \nPO-Revision-Date: 2021-04-08 08:14-0500\nLanguage-Team: \nMIME-Version: 1.0\nContent-Type: text/plain; charset=UTF-8\nContent-Transfer-Encoding: 8bit\nX-Generator: Poedit 2.4.2\nLast-Translator: \nPlural-Forms: nplurals=2; plural=(n != 1);\nLanguage: es_CO\n\x00Tengo napari\x00Tengo naparis\x00Tengo {n} napari con {variable}\x00Tengo {n} naparis con {variable}\x00M\xc3\xa1s sobre napari\x00M\xc3\xa1s sobre napari con {variable}\x00Tengo napari con contexto\x00Tengo naparis con contexto\x00Tengo {n} napari con {variable} y contexto\x00Tengo {n} naparis con {variable} y contexto\x00M\xc3\xa1s sobre napari con contexto\x00M\xc3\xa1s sobre napari con contexto y {variable}\x00'


@pytest.fixture
def trans(tmp_path):
    """A good plugin that uses entry points."""
    distinfo = tmp_path / 'napari_language_pack_es_CO-0.1.0.dist-info'
    distinfo.mkdir()
    (distinfo / 'top_level.txt').write_text('napari_language_pack_es_CO')
    (distinfo / 'entry_points.txt').write_text(
        '[napari.languagepack]\nes_CO = napari_language_pack_es_CO\n'
    )
    (distinfo / 'METADATA').write_text(
        'Metadata-Version: 2.1\n'
        'Name: napari-language-pack-es-CO\n'
        'Version: 0.1.0\n'
    )
    pkgdir = tmp_path / 'napari_language_pack_es_CO'
    msgs = pkgdir / 'locale' / 'es_CO' / 'LC_MESSAGES'
    msgs.mkdir(parents=True)
    (pkgdir / '__init__.py').touch()
    (msgs / 'napari.po').write_text(es_CO_po)
    (msgs / 'napari.mo').write_bytes(es_CO_mo)

    from napari_plugin_engine.manager import temp_path_additions

    with temp_path_additions(tmp_path):
        # Load translator and force a locale for testing
        translator._set_locale(TEST_LOCALE)
        return translator.load()


def test_get_language_packs(trans):
    result = get_language_packs()
    assert result == {
        'en': {'displayName': 'English', 'nativeName': 'English'}
    }


def test_get_display_name_valid():
    assert _get_display_name('en', 'en') == 'English'
    assert _get_display_name('en', 'es') == 'Inglés'
    assert _get_display_name('en', 'es_CO') == 'Inglés'
    assert _get_display_name('en', 'fr') == 'Anglais'
    assert _get_display_name('es', 'en') == 'Spanish'
    assert _get_display_name('fr', 'en') == 'French'


def test_get_display_name_invalid():
    assert _get_display_name('en', 'foo') == 'English'
    assert _get_display_name('foo', 'en') == 'English'
    assert _get_display_name('foo', 'bar') == 'English'


def test_is_valid_locale_valid():
    assert _is_valid_locale('en')
    assert _is_valid_locale('es')
    assert _is_valid_locale('es_CO')


def test_is_valid_locale_invalid():
    assert not _is_valid_locale('foo_SPAM')
    assert not _is_valid_locale('bar')


def test_load_language_valid(tmp_path):
    # This is valid content
    data = """
application:
  language: es_ES
"""
    temp_config_path = tmp_path / 'tempconfig.yml'
    with open(temp_config_path, 'w') as fh:
        fh.write(data)

    result = _load_language(temp_config_path)
    assert result == 'es_ES'


def test_load_language_invalid(tmp_path):
    # This is invalid content
    data = ':'
    temp_config_path = tmp_path / 'tempconfig.yml'
    with open(temp_config_path, 'w') as fh:
        fh.write(data)

    with pytest.warns(UserWarning):
        _load_language(temp_config_path)


def test_locale_invalid():
    with pytest.warns(UserWarning):
        translator._set_locale(TEST_LOCALE)
        trans = translator.load()
        result = trans._('BOO')
        assert result == 'BOO'


# Test trans methods
# ------------------
def test_locale_singular(trans):
    expected_result = 'Más sobre napari'
    result = trans._('More about napari')
    assert result == expected_result


def test_locale_singular_with_format(trans):
    variable = 1
    singular = 'More about napari with {variable}'
    expected_result = f'Más sobre napari con {variable}'
    result = trans._(singular, variable=variable)
    assert result == expected_result


def test_locale_singular_deferred_with_format(trans):
    variable = 1
    singular = 'More about napari with {variable}'
    original_result = f'More about napari with {variable}'
    translated_result = f'Más sobre napari con {variable}'
    result = trans._(singular, deferred=True, variable=variable)
    assert isinstance(result, TranslationString)
    assert result.translation() == translated_result
    assert result.value() == original_result
    assert str(result) == original_result


def test_locale_singular_context(trans):
    context = 'singular-context'
    singular = 'More about napari with context'

    result = trans._p(context, singular)
    assert result == 'Más sobre napari con contexto'


def test_locale_singular_context_with_format(trans):
    context = 'singular-context-variables'
    variable = 1
    singular = 'More about napari with context and {variable}'

    result = trans._p(context, singular, variable=variable)
    assert result == f'Más sobre napari con contexto y {variable}'


def test_locale_singular_context_deferred_with_format(trans):
    context = 'singular-context-variables'
    variable = 1
    singular = 'More about napari with context and {variable}'
    original_result = f'More about napari with context and {variable}'

    translated_result = f'Más sobre napari con contexto y {variable}'

    result = trans._p(context, singular, deferred=True, variable=variable)
    assert isinstance(result, TranslationString)
    assert result.translation() == translated_result
    assert result.value() == original_result
    assert str(result) == original_result


def test_locale_plural(trans):
    singular = 'I have napari'
    plural = 'I have naparis'

    n = 1
    result = trans._n(singular, plural, n=n)
    assert result == 'Tengo napari'

    n = 2
    result_plural = trans._n(singular, plural, n=n)
    assert result_plural == 'Tengo naparis'


def test_locale_plural_with_format(trans):
    singular = 'I have {n} napari with {variable}'
    plural = 'I have {n} naparis with {variable}'
    variable = 1

    n = 1
    result = trans._n(singular, plural, n=n, variable=variable)
    expected_result = f'Tengo {n} napari con {variable}'
    assert result == expected_result

    n = 2
    result_plural = trans._n(singular, plural, n=n, variable=variable)
    expected_result_plural = f'Tengo {n} naparis con {variable}'
    assert result_plural == expected_result_plural


def test_locale_plural_deferred_with_format(trans):
    variable = 1
    singular = 'I have {n} napari with {variable}'
    plural = 'I have {n} naparis with {variable}'

    n = 1
    original_result = singular.format(n=n, variable=variable)
    result = trans._n(singular, plural, n=n, deferred=True, variable=variable)
    expected_result = f'Tengo {n} napari con {variable}'
    assert isinstance(result, TranslationString)
    assert result.translation() == expected_result
    assert result.value() == original_result
    assert str(result) == original_result

    n = 2
    original_result_plural = plural.format(n=n, variable=variable)
    result_plural = trans._n(
        singular, plural, n=n, deferred=True, variable=variable
    )
    expected_result_plural = f'Tengo {n} naparis con {variable}'
    assert isinstance(result, TranslationString)
    assert result_plural.translation() == expected_result_plural
    assert result_plural.value() == original_result_plural
    assert str(result_plural) == original_result_plural


def test_locale_plural_context(trans):
    context = 'plural-context'
    singular = 'I have napari with context'
    plural = 'I have naparis with context'

    n = 1
    result = trans._np(context, singular, plural, n=n)
    assert result == 'Tengo napari con contexto'

    n = 2
    result_plural = trans._np(context, singular, plural, n=n)
    assert result_plural == 'Tengo naparis con contexto'


def test_locale_plural_context_with_format(trans):
    context = 'plural-context-variables'
    singular = 'I have {n} napari with {variable} and context'
    plural = 'I have {n} naparis with {variable} and context'
    variable = 1

    n = 1
    result = trans._np(context, singular, plural, n=n, variable=variable)
    assert result == f'Tengo {n} napari con {variable} y contexto'

    n = 2
    result_plural = trans._np(
        context, singular, plural, n=n, variable=variable
    )
    assert result_plural == f'Tengo {n} naparis con {variable} y contexto'


def test_locale_plural_context_deferred_with_format(trans):
    context = 'plural-context-variables'
    variable = 1
    singular = 'I have {n} napari with {variable} and context'
    plural = 'I have {n} naparis with {variable} and context'

    n = 1
    original_result = singular.format(n=n, variable=variable)
    result = trans._np(
        context, singular, plural, n=n, deferred=True, variable=variable
    )
    expected_result = f'Tengo {n} napari con {variable} y contexto'

    assert isinstance(result, TranslationString)
    assert result.translation() == expected_result
    assert result.value() == original_result
    assert str(result) == original_result

    n = 2
    original_result_plural = plural.format(n=n, variable=variable)
    result_plural = trans._np(
        context, singular, plural, n=n, deferred=True, variable=variable
    )
    expected_result_plural = f'Tengo {n} naparis con {variable} y contexto'

    assert isinstance(result, TranslationString)
    assert result_plural.translation() == expected_result_plural
    assert result_plural.value() == original_result_plural
    assert str(result_plural) == original_result_plural


# Deferred strings in exceptions
# ------------------------------
def test_exception_string(trans):
    expected_result = 'Más sobre napari'
    result = trans._('MORE ABOUT NAPARI', deferred=True)
    assert str(result) != expected_result
    assert str(result) == 'MORE ABOUT NAPARI'

    with pytest.raises(ValueError, match='MORE ABOUT NAPARI') as err:
        raise ValueError(result)

    assert isinstance(err.value.args[0], TranslationString)


# Test TranslationString
# ----------------------
def test_translation_string_exceptions():
    with pytest.raises(ValueError, match='Must provide at least'):
        TranslationString()


def test_bundle_exceptions(trans):
    with pytest.raises(
        TypeError, match="missing 1 required keyword-only argument: 'msgid'"
    ):
        trans._dnpgettext()


@pytest.mark.parametrize(
    'kwargs',
    [
        {
            'msgid': 'huhu',
        },
        {
            'msgid': 'Convert to {dtype}',
            'dtype': 'uint16',
        },
        {
            'msgid': 'Convert to {dtype}',
            'dtype': 'uint16',
            'deferred': True,
        },
        {
            'msgid': 'Convert to {dtype}',
            'dtype': 'uint16',
            'deferred': False,
        },
        {
            'msgid': 'Convert to {dtype}',
            'msgid_plural': 'Convert to {dtype}s',
            'n': 1,
            'dtype': 'uint16',
        },
        {
            'msgid': 'Convert to {dtype}',
            'msgid_plural': 'Convert to {dtype}s',
            'n': 2,
            'dtype': 'uint16',
        },
    ],
)
def test_deepcopy(kwargs):
    """Object containing translation strings can't be deep-copied.

    See:
      - https://github.com/napari/napari/issues/2911
      - https://github.com/napari/napari/issues/4736
    """
    kwargs['domain'] = 'napari'
    t = TranslationString(**kwargs)
    u = deepcopy(t)
    assert t is not u
    assert t == u



from napari._vispy.layers.tracks import VispyTracksLayer
from napari.layers import Tracks


def test_tracks_graph_cleanup():
    """
    Test if graph data can be cleaned up without any issue.
    There was problems with the shader buffer once, see issue #4155.
    """
    tracks_data = [
        [1, 0, 236, 0],
        [1, 1, 236, 100],
        [1, 2, 236, 200],
        [2, 3, 436, 500],
        [2, 4, 436, 1000],
        [3, 3, 636, 500],
        [3, 4, 636, 1000],
    ]
    graph = {1: [], 2: [1], 3: [1]}

    layer = Tracks(tracks_data, graph=graph)
    visual = VispyTracksLayer(layer)

    layer.graph = {}

    assert visual.node._subvisuals[2]._pos is None
    assert visual.node._subvisuals[2]._connect is None



import numpy as np
import pytest

from napari._vispy.utils.gl import fix_data_dtype

BUILTINS_DISP = 'napari'
BUILTINS_NAME = 'builtins'


# Previously tests often segfaulted on CI at the 26th test of test_qt_viewer
# That test (number 26) was split off to make debugging easier
# See https://github.com/napari/napari/pull/5676
@pytest.mark.parametrize(
    'dtype',
    [
        'int8',
        'uint8',
        'int16',
        'uint16',
        'int32',
        'float16',
        'float32',
        'float64',
    ],
)
def test_qt_viewer_data_integrity(make_napari_viewer, dtype):
    """Test that the viewer doesn't change the underlying array."""
    image = np.random.rand(10, 32, 32)
    image *= 200 if dtype.endswith('8') else 2**14
    image = image.astype(dtype)
    imean = image.mean()

    viewer = make_napari_viewer()
    layer = viewer.add_image(image.copy())
    data = layer.data

    datamean = np.mean(data)
    assert datamean == imean
    # toggle dimensions
    viewer.dims.ndisplay = 3
    datamean = np.mean(data)
    assert datamean == imean
    # back to 2D
    viewer.dims.ndisplay = 2
    datamean = np.mean(data)
    assert datamean == imean
    # also check that vispy gets (almost) the same data
    datamean = np.mean(fix_data_dtype(data))
    assert np.allclose(datamean, imean, rtol=5e-04)


@pytest.mark.parametrize(
    ('dtype', 'expected'),
    [
        (np.bool_, np.uint8),
        (np.int8, np.float32),
        (np.uint8, np.uint8),
        (np.int16, np.float32),
        (np.uint16, np.uint16),
        (np.uint32, np.float32),
        (np.float32, np.float32),
        (np.float64, np.float32),
    ],
)
def test_fix_data_dtype_big_values(dtype, expected):
    data = np.array([0, 2, 2**17], dtype=np.int32).astype(dtype)
    casted = fix_data_dtype(data)
    assert np.allclose(casted, data)
    assert casted.dtype == expected



import sys

import numpy.testing as npt
import pyautogui
import pytest
from qtpy.QtCore import QPoint, Qt
from qtpy.QtWidgets import QApplication

from napari._pydantic_compat import BaseModel
from napari._qt.dialogs.preferences_dialog import (
    PreferencesDialog,
    QMessageBox,
)
from napari._tests.utils import skip_local_focus, skip_on_mac_ci
from napari._vendor.qt_json_builder.qt_jsonschema_form.widgets import (
    EnumSchemaWidget,
    FontSizeSchemaWidget,
    HighlightPreviewWidget,
    HorizontalObjectSchemaWidget,
)
from napari.settings import NapariSettings, get_settings
from napari.settings._constants import BrushSizeOnMouseModifiers, LabelDTypes
from napari.utils.interactions import Shortcut
from napari.utils.key_bindings import KeyBinding


@pytest.fixture
def pref(qtbot):
    dlg = PreferencesDialog()
    qtbot.addWidget(dlg)
    # check settings default values and change them for later checks
    settings = get_settings()
    # change theme setting (default `dark`)
    assert settings.appearance.theme == 'dark'
    dlg._settings.appearance.theme = 'light'
    assert get_settings().appearance.theme == 'light'
    # change highlight setting related value (default thickness `1`)
    assert get_settings().appearance.highlight.highlight_thickness == 1
    dlg._settings.appearance.highlight.highlight_thickness = 5
    assert get_settings().appearance.highlight.highlight_thickness == 5
    # change `napari:reset_scroll_progress` shortcut/keybinding (default keybinding `Ctrl`/`Control`)
    # a copy of the initial `shortcuts` dictionary needs to be done since, to trigger an
    # event update from the `ShortcutsSettings` model, the whole `shortcuts` dictionary
    # needs to be reassigned.
    assert dlg._settings.shortcuts.shortcuts[
        'napari:reset_scroll_progress'
    ] == [KeyBinding.from_str('Ctrl')]
    shortcuts = dlg._settings.shortcuts.shortcuts.copy()
    shortcuts['napari:reset_scroll_progress'] = [KeyBinding.from_str('U')]
    dlg._settings.shortcuts.shortcuts = shortcuts
    assert dlg._settings.shortcuts.shortcuts[
        'napari:reset_scroll_progress'
    ] == [KeyBinding.from_str('U')]
    return dlg


def test_prefdialog_populated(pref):
    subfields = filter(
        lambda f: isinstance(f.type_, type) and issubclass(f.type_, BaseModel),
        NapariSettings.__fields__.values(),
    )
    assert pref._stack.count() == len(list(subfields))


def test_dask_widget(qtbot, pref):
    dask_widget = pref._stack.currentWidget().widget().widget.widgets['dask']
    def_dask_enabled = True
    settings = pref._settings

    # check custom widget definition and default value for dask cache `enabled` setting
    assert isinstance(dask_widget, HorizontalObjectSchemaWidget)
    assert settings.application.dask.enabled == def_dask_enabled
    assert dask_widget.state['enabled'] == def_dask_enabled

    # check changing dask cache `enabled` setting via widget
    new_dask_enabled = False
    dask_widget.state = {
        'enabled': new_dask_enabled,
        'cache': dask_widget.state['cache'],
    }
    assert settings.application.dask.enabled == new_dask_enabled
    assert dask_widget.state['enabled'] == new_dask_enabled
    assert dask_widget.widgets['enabled'].state == new_dask_enabled

    # check changing dask `enabled` setting via settings object (to default value)
    settings.application.dask.enabled = def_dask_enabled
    assert dask_widget.state['enabled'] == def_dask_enabled
    assert dask_widget.widgets['enabled'].state == def_dask_enabled


def test_font_size_widget(qtbot, pref):
    font_size_widget = (
        pref._stack.widget(1).widget().widget.widgets['font_size']
    )
    def_font_size = 12 if sys.platform == 'darwin' else 9

    # check custom widget definition usage for the font size setting
    # and default values
    assert isinstance(font_size_widget, FontSizeSchemaWidget)
    assert get_settings().appearance.font_size == def_font_size
    assert font_size_widget.state == def_font_size

    # check setting a new font size value via widget
    new_font_size = 14
    font_size_widget.state = new_font_size
    assert get_settings().appearance.font_size == new_font_size

    # verify that a theme change preserves the font size value
    assert get_settings().appearance.theme == 'light'
    get_settings().appearance.theme = 'dark'
    assert get_settings().appearance.font_size == new_font_size
    assert font_size_widget.state == new_font_size

    # check reset button works
    font_size_widget._reset_button.click()
    assert get_settings().appearance.font_size == def_font_size
    assert font_size_widget.state == def_font_size


@pytest.mark.parametrize(
    ('enum_setting_name', 'enum_setting_class'),
    [
        ('new_labels_dtype', LabelDTypes),
        ('brush_size_on_mouse_move_modifiers', BrushSizeOnMouseModifiers),
    ],
)
def test_StrEnum_widgets(qtbot, pref, enum_setting_name, enum_setting_class):
    enum_widget = (
        pref._stack.currentWidget().widget().widget.widgets[enum_setting_name]
    )
    settings = pref._settings

    # check custom widget definition and widget value follows setting
    assert isinstance(enum_widget, EnumSchemaWidget)
    assert enum_widget.state == getattr(
        settings.application, enum_setting_name
    )

    # check changing setting via widget
    for idx in range(enum_widget.count()):
        item_text = enum_widget.itemText(idx)
        item_data = enum_widget.itemData(idx)
        enum_widget.setCurrentText(item_text)
        assert getattr(settings.application, enum_setting_name) == item_data
        assert enum_widget.state == item_data

    # check changing setting updates widget
    for enum_value in enum_setting_class:
        setattr(settings.application, enum_setting_name, enum_value)
        assert enum_widget.state == enum_value


def test_highlight_widget(qtbot, pref):
    highlight_widget = (
        pref._stack.widget(1).widget().widget.widgets['highlight']
    )
    settings = pref._settings

    # check custom widget definition and widget follows settings values
    assert isinstance(highlight_widget, HighlightPreviewWidget)
    assert (
        highlight_widget.state['highlight_color']
        == settings.appearance.highlight.highlight_color
    )
    assert (
        highlight_widget.state['highlight_thickness']
        == settings.appearance.highlight.highlight_thickness
    )

    # check changing setting via widget
    new_widget_values = {
        'highlight_thickness': 5,
        'highlight_color': [0.6, 0.6, 1.0, 1.0],
    }
    highlight_widget.setValue(new_widget_values)
    npt.assert_allclose(
        settings.appearance.highlight.highlight_color,
        new_widget_values['highlight_color'],
    )
    assert (
        settings.appearance.highlight.highlight_thickness
        == new_widget_values['highlight_thickness']
    )

    # check changing setting updates widget
    new_setting_values = {
        'highlight_thickness': 1,
        'highlight_color': [0.5, 0.6, 1.0, 1.0],
    }

    settings.appearance.highlight.highlight_color = new_setting_values[
        'highlight_color'
    ]
    npt.assert_allclose(
        highlight_widget.state['highlight_color'],
        new_setting_values['highlight_color'],
    )

    settings.appearance.highlight.highlight_thickness = new_setting_values[
        'highlight_thickness'
    ]
    assert (
        highlight_widget.state['highlight_thickness']
        == new_setting_values['highlight_thickness']
    )


def test_preferences_dialog_accept(qtbot, pref):
    with qtbot.waitSignal(pref.finished):
        pref.accept()
    assert get_settings().appearance.theme == 'light'


def test_preferences_dialog_ok(qtbot, pref):
    with qtbot.waitSignal(pref.finished):
        pref._button_ok.click()
    assert get_settings().appearance.theme == 'light'


def test_preferences_dialog_close(qtbot, pref):
    with qtbot.waitSignal(pref.finished):
        pref.close()
    assert get_settings().appearance.theme == 'light'


def test_preferences_dialog_escape(qtbot, pref):
    with qtbot.waitSignal(pref.finished):
        qtbot.keyPress(pref, Qt.Key_Escape)
    assert get_settings().appearance.theme == 'light'


@pytest.mark.key_bindings
def test_preferences_dialog_cancel(qtbot, pref):
    with qtbot.waitSignal(pref.finished):
        pref._button_cancel.click()
    assert get_settings().appearance.theme == 'dark'
    assert get_settings().shortcuts.shortcuts[
        'napari:reset_scroll_progress'
    ] == [KeyBinding.from_str('Ctrl')]


@pytest.mark.key_bindings
def test_preferences_dialog_restore(qtbot, pref, monkeypatch):
    theme_widget = pref._stack.widget(1).widget().widget.widgets['theme']
    highlight_widget = (
        pref._stack.widget(1).widget().widget.widgets['highlight']
    )
    shortcut_widget = (
        pref._stack.widget(3).widget().widget.widgets['shortcuts']
    )

    assert get_settings().appearance.theme == 'light'
    assert theme_widget.state == 'light'
    assert get_settings().appearance.highlight.highlight_thickness == 5
    assert highlight_widget.state['highlight_thickness'] == 5
    assert get_settings().shortcuts.shortcuts[
        'napari:reset_scroll_progress'
    ] == [KeyBinding.from_str('U')]
    assert KeyBinding.from_str(
        Shortcut.parse_platform(
            shortcut_widget._table.item(
                0, shortcut_widget._shortcut_col
            ).text()
        )
    ) == KeyBinding.from_str('U')

    monkeypatch.setattr(
        QMessageBox, 'question', lambda *a: QMessageBox.RestoreDefaults
    )
    pref._restore_default_dialog()

    assert get_settings().appearance.theme == 'dark'
    assert theme_widget.state == 'dark'
    assert get_settings().appearance.highlight.highlight_thickness == 1
    assert highlight_widget.state['highlight_thickness'] == 1
    assert get_settings().shortcuts.shortcuts[
        'napari:reset_scroll_progress'
    ] == [KeyBinding.from_str('Ctrl')]
    assert KeyBinding.from_str(
        Shortcut.parse_platform(
            shortcut_widget._table.item(
                0, shortcut_widget._shortcut_col
            ).text()
        )
    ) == KeyBinding.from_str('Ctrl')


@skip_local_focus
@skip_on_mac_ci
@pytest.mark.key_bindings
@pytest.mark.parametrize(
    'confirm_key',
    ['enter', 'return', 'tab'],
)
def test_preferences_dialog_not_dismissed_by_keybind_confirm(
    qtbot, pref, confirm_key
):
    """This test ensures that when confirming a keybinding change, the dialog is not dismissed.

    Notes:
        * Skipped on macOS CI due to accessibility permissions not being
          settable on macOS GitHub Actions runners.
        * For this test to pass locally, you need to give the Terminal/iTerm/VSCode
          application accessibility permissions:
              `System Settings > Privacy & Security > Accessibility`

        See https://github.com/asweigart/pyautogui/issues/247 and
        https://github.com/asweigart/pyautogui/issues/247#issuecomment-437668855
    """
    shortcut_widget = (
        pref._stack.widget(3).widget().widget.widgets['shortcuts']
    )
    pref._stack.setCurrentIndex(3)
    # ensure the dialog is showing
    pref.show()
    qtbot.waitExposed(pref)
    assert pref.isVisible()

    shortcut = shortcut_widget._table.item(
        0, shortcut_widget._shortcut_col
    ).text()
    assert shortcut == 'U'

    x = shortcut_widget._table.columnViewportPosition(
        shortcut_widget._shortcut_col
    )
    y = shortcut_widget._table.rowViewportPosition(0)

    item_pos = QPoint(x, y)
    qtbot.mouseClick(
        shortcut_widget._table.viewport(),
        Qt.MouseButton.LeftButton,
        pos=item_pos,
    )
    qtbot.mouseDClick(
        shortcut_widget._table.viewport(),
        Qt.MouseButton.LeftButton,
        pos=item_pos,
    )
    qtbot.waitUntil(lambda: QApplication.focusWidget() is not None)
    pyautogui.press('delete')
    qtbot.wait(100)
    pyautogui.press(confirm_key)
    qtbot.wait(100)

    # ensure the dialog is still open
    assert pref.isVisible()

    # verify that the keybind is changed
    shortcut = shortcut_widget._table.item(
        0, shortcut_widget._shortcut_col
    ).text()
    assert shortcut == ''



import numpy as np
import numpy.typing as npt


def elements_in_slice(
    index: tuple[npt.NDArray[np.int_], ...], position_in_axes: dict[int, int]
) -> npt.NDArray[np.bool_]:
    """Mask elements from a multi-dimensional index not in a given slice.

    Some n-D operations may edit data that is not visible in the current slice.
    Given slice position information (as a dictionary mapping axis to index on that
    axis), this function returns a boolean mask for the possibly higher-dimensional
    multi-index so that elements not currently visible are masked out. The
    resulting multi-index can then be subset and used to index into a texture or
    other lower-dimensional view.

    Parameters
    ----------
    index : tuple of array of int
        A NumPy fancy indexing expression [1]_.
    position_in_axes : dict[int, int]
        A dictionary mapping sliced (non-displayed) axes to a slice position.

    Returns
    -------
    visible : array of bool
        A boolean array indicating which items are visible in the current view.
    """
    queries = [
        index[ax] == position for ax, position in position_in_axes.items()
    ]
    return np.logical_and.reduce(queries, axis=0)


def index_in_slice(
    index: tuple[npt.NDArray[np.int_], ...],
    position_in_axes: dict[int, int],
    indices_order: tuple[int, ...],
) -> tuple[npt.NDArray[np.int_], ...]:
    """Convert a NumPy fancy indexing expression from data to sliced space.

    Parameters
    ----------
    index : tuple of array of int
        A NumPy fancy indexing expression [1]_.
    position_in_axes : dict[int, int]
        A dictionary mapping sliced (non-displayed) axes to a slice position.
    indices_order : tuple of int
        The order of the indices in data view.

    Returns
    -------
    sliced_index : tuple of array of int
        The indexing expression (nD) restricted to the current slice (usually
        2D or 3D).

    Examples
    --------
    >>> index = (np.arange(5), np.full(5, 1), np.arange(4, 9))
    >>> index_in_slice(index, {0: 3}, (0, 1, 2))
    (array([1]), array([7]))
    >>> index_in_slice(index, {1: 1, 2: 8}, (0, 1, 2))
    (array([4]),)

    References
    ----------
    [1]: https://numpy.org/doc/stable/user/basics.indexing.html#integer-array-indexing
    """
    index_in_slice = elements_in_slice(index, position_in_axes)
    return tuple(
        index[i][index_in_slice]
        for i in indices_order
        if i not in position_in_axes
    )



from qtpy.QtCore import Signal
from qtpy.QtWidgets import QHBoxLayout, QPushButton, QWidget

from napari._qt.widgets.qt_spinbox import QtSpinBox
from napari.settings import get_settings
from napari.utils.theme import get_system_theme, get_theme
from napari.utils.translations import trans


class QtFontSizeWidget(QWidget):
    """
    Widget to change `font_size` and enable to reset is value to the current
    selected theme default `font_size` value.
    """

    valueChanged = Signal(int)

    def __init__(self, parent: QWidget = None) -> None:
        super().__init__(parent=parent)
        self._spinbox = QtSpinBox()
        self._reset_button = QPushButton(trans._('Reset font size'))

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._spinbox)
        layout.addWidget(self._reset_button)
        self.setLayout(layout)

        self._spinbox.valueChanged.connect(self.valueChanged)
        self._reset_button.clicked.connect(self._reset)

    def _reset(self) -> None:
        """
        Reset the widget value to the current selected theme font size value.
        """
        current_theme_name = get_settings().appearance.theme
        if current_theme_name == 'system':
            # system isn't a theme, so get the name
            current_theme_name = get_system_theme()
        current_theme = get_theme(current_theme_name)
        self.setValue(int(current_theme.font_size[:-2]))

    def value(self) -> int:
        """
        Return the current widget value.

        Returns
        -------
        int
            The current value.
        """
        return self._spinbox.value()

    def setValue(self, value: int) -> None:
        """
        Set the current widget value.

        Parameters
        ----------
        value : int
            The current value.
        """
        self._spinbox.setValue(value)

    def setRange(self, min_value: int, max_value: int) -> None:
        """
        Value range that the spinbox widget will use.

        Parameters
        ----------
        min_value : int
            Minimum value the font_size could be set.
        max_value : int
            Maximum value the font_size could be set.
        """
        self._spinbox.setRange(min_value, max_value)



from functools import lru_cache

from npe2 import (
    PluginManager as _PluginManager,
)

from napari.plugins import _npe2
from napari.plugins._plugin_manager import NapariPluginManager
from napari.settings import get_settings

__all__ = ('menu_item_template', 'plugin_manager')

from napari.utils.theme import _install_npe2_themes

#: Template to use for namespacing a plugin item in the menu bar
# widget_name (plugin_name)
menu_item_template = '{1} ({0})'
"""Template to use for namespacing a plugin item in the menu bar"""
#: The main plugin manager instance for the `napari` plugin namespace.
plugin_manager = NapariPluginManager()
"""Main Plugin manager instance"""


@lru_cache  # only call once
def _initialize_plugins() -> None:
    _npe2pm = _PluginManager.instance()

    settings = get_settings()
    if settings.schema_version >= '0.4.0':
        for p in settings.plugins.disabled_plugins:
            _npe2pm.disable(p)

    # just in case anything has already been registered before we initialized
    _npe2.on_plugins_registered(set(_npe2pm.iter_manifests()))

    # connect enablement/registration events to listeners
    _npe2pm.events.enablement_changed.connect(
        _npe2.on_plugin_enablement_change
    )
    _npe2pm.events.plugins_registered.connect(_npe2.on_plugins_registered)
    _npe2pm.discover(include_npe1=settings.plugins.use_npe2_adaptor)

    # Disable plugins listed as disabled in settings, or detected in npe2
    _from_npe2 = {m.name for m in _npe2pm.iter_manifests()}
    if 'napari' in _from_npe2:
        _from_npe2.update({'napari', 'builtins'})
    plugin_manager._skip_packages = _from_npe2
    plugin_manager._blocked.update(settings.plugins.disabled_plugins)

    if settings.plugins.use_npe2_adaptor:
        # prevent npe1 plugin_manager discovery
        # (this doesn't prevent manual registration)
        plugin_manager.discover = lambda *a, **k: None
    else:
        plugin_manager._initialize()

    _install_npe2_themes()



from dataclasses import dataclass
from unittest.mock import patch

import numpy as np
import pytest

from napari.components.viewer_model import ViewerModel
from napari.utils._proxies import PublicOnlyProxy, ReadOnlyWrapper
from napari.utils.events.containers._set import EventedSet


def test_ReadOnlyWrapper_setitem():
    """test that ReadOnlyWrapper prevents setting items"""
    d = {'hi': 3}
    d_read_only = ReadOnlyWrapper(d)

    with pytest.raises(TypeError):
        d_read_only['hi'] = 5


def test_ReadOnlyWrapper_setattr():
    """test that ReadOnlyWrapper prevents setting attributes"""

    class TestClass:
        x = 3

    tc = TestClass()
    tc_read_only = ReadOnlyWrapper(tc)

    with pytest.raises(TypeError):
        tc_read_only.x = 5


@pytest.fixture
def _patched_root_dir():
    """Simulate a call from outside of napari"""
    with patch('napari.utils.misc.ROOT_DIR', new='/some/other/package'):
        yield


@pytest.mark.usefixtures('_patched_root_dir')
def test_PublicOnlyProxy():
    class X:
        a = 1
        _b = 'nope'

        def method(self):
            return 2

    class Tester:
        x = X()
        _private = 2

        def __getitem__(self, key):
            return X()

    t = Tester()
    proxy = PublicOnlyProxy(t)
    assert proxy.x.a == 1
    assert proxy[0].a == 1
    assert proxy.x.method() == 2

    assert isinstance(proxy, Tester)

    with pytest.warns(FutureWarning, match='Private attribute access'):
        proxy._private

    with pytest.warns(FutureWarning, match='Private attribute access'):
        # warns on setattr
        proxy._private = 4

    with pytest.warns(FutureWarning, match='Private attribute access'):
        # works on sub-objects too
        proxy.x._b

    with pytest.warns(FutureWarning, match='Private attribute access'):
        # works on sub-items too
        proxy[0]._b

    assert '_private' not in dir(proxy)
    assert '_private' in dir(t)


@pytest.mark.filterwarnings('ignore:Qt libs are available but')
def test_thread_proxy_guard(monkeypatch, single_threaded_executor):
    class X:
        a = 1

    monkeypatch.setenv('NAPARI_ENSURE_PLUGIN_MAIN_THREAD', 'True')

    x = X()
    x_proxy = PublicOnlyProxy(x)

    f = single_threaded_executor.submit(x.__setattr__, 'a', 2)
    f.result()
    assert x.a == 2

    f = single_threaded_executor.submit(x_proxy.__setattr__, 'a', 3)
    with pytest.raises(RuntimeError):
        f.result()
    assert x.a == 2


@pytest.mark.usefixtures('_patched_root_dir')
def test_public_proxy_limited_to_napari():
    """Test that the recursive public proxy goes no farther than napari."""
    viewer = ViewerModel()
    viewer.add_points(None)
    pv = PublicOnlyProxy(viewer)
    assert not isinstance(pv.layers[0].data, PublicOnlyProxy)


@pytest.mark.usefixtures('_patched_root_dir')
def test_array_from_proxy_objects():
    """Test that the recursive public proxy goes no farther than napari."""
    viewer = ViewerModel()
    viewer.add_points(None)
    pv = PublicOnlyProxy(viewer)
    assert isinstance(np.array(pv.dims.displayed, dtype=int), np.ndarray)


def test_receive_return_proxy_object():
    """Test that an"""
    viewer = ViewerModel()
    viewer.add_image(np.random.random((20, 20)))
    pv = PublicOnlyProxy(viewer)

    # simulates behavior from outside of napari
    with patch('napari.utils.misc.ROOT_DIR', new='/some/other/package'):
        # the recursion means this layer will be a Proxy Object on __getitem__
        layer = pv.layers[-1]
        assert isinstance(layer, PublicOnlyProxy)
        # remove and add it back, should be fine
        add_layer = pv.add_layer
        viewer.layers.pop()

    add_layer(layer)
    assert len(viewer.layers) == 1


def test_viewer_method():
    viewer = PublicOnlyProxy(ViewerModel())
    assert viewer.add_points() is not None


def test_unwrap_on_call():
    """Check that PublicOnlyProxy'd arguments to methods of a
    PublicOnlyProxy'd object are unwrapped before calling the method.
    """
    evset = EventedSet()
    public_only_evset = PublicOnlyProxy(evset)
    text = 'aaa'
    wrapped_text = PublicOnlyProxy(text)
    public_only_evset.add(wrapped_text)
    retrieved_text = next(iter(evset))

    # check that the text in the set is not the version wrapped with
    # PublicOnlyProxy
    assert id(text) == id(retrieved_text)


def test_unwrap_setattr():
    """Check that objects added with __setattr__ of an object wrapped with
    PublicOnlyProxy are unwrapped before setting the attribute.
    """

    @dataclass
    class Sample:
        attribute = 'aaa'

    sample = Sample()
    public_only_sample = PublicOnlyProxy(sample)

    text = 'bbb'
    wrapped_text = PublicOnlyProxy(text)

    public_only_sample.attribute = wrapped_text
    attribute = sample.attribute  # use original, not wrapped object

    # check that the attribute in the unwrapped sample is itself not the
    # wrapped text, but the original text.
    assert id(text) == id(attribute)



from napari.components.overlays.brush_circle import BrushCircleOverlay


def test_brush_circle():
    """Test creating a brush circle overlay"""
    brush_circle = BrushCircleOverlay()
    assert brush_circle is not None



import typing

import numpy as np

from napari._vispy.layers.base import VispyBaseLayer
from napari._vispy.utils.gl import BLENDING_MODES
from napari._vispy.utils.text import update_text
from napari._vispy.visuals.shapes import ShapesVisual
from napari.settings import get_settings
from napari.utils.events import disconnect_events

if typing.TYPE_CHECKING:
    from napari.layers import Shapes


class VispyShapesLayer(VispyBaseLayer):
    node: ShapesVisual
    layer: 'Shapes'

    def __init__(self, layer: 'Shapes') -> None:
        node = ShapesVisual()
        super().__init__(layer, node)

        self.layer.events.edge_width.connect(self._on_data_change)
        self.layer.events.edge_color.connect(self._on_data_change)
        self.layer.events.face_color.connect(self._on_data_change)
        self.layer.events.highlight.connect(self._on_highlight_change)
        self.layer.text.events.connect(self._on_text_change)
        self.layer.events.scale_factor.connect(self._update_text)

        # TODO: move to overlays
        self.node.highlight_vertices.symbol = 'square'
        self.node.highlight_vertices.scaling = False

        self.reset()
        self._on_data_change()

    def _on_data_change(self):
        faces = self.layer._data_view._mesh.displayed_triangles
        colors = self.layer._data_view._mesh.displayed_triangles_colors
        vertices = self.layer._data_view._mesh.vertices

        # Note that the indices of the vertices need to be reversed to
        # go from numpy style to xyz
        if vertices is not None:
            vertices = vertices[:, ::-1]

        if len(vertices) == 0 or len(faces) == 0:
            vertices = np.zeros((3, self.layer._slice_input.ndisplay))
            faces = np.array([[0, 1, 2]])
            colors = np.array([[0, 0, 0, 0]])

        if (
            len(self.layer.data)
            and self.layer._slice_input.ndisplay == 3
            and self.layer.ndim == 2
        ):
            vertices = np.pad(vertices, ((0, 0), (0, 1)), mode='constant')

        self.node.shape_faces.set_data(
            vertices=vertices, faces=faces, face_colors=colors
        )

        # Call to update order of translation values with new dims:
        self._on_matrix_change()
        self._update_text(update_node=False)
        self.node.update()

    def _on_highlight_change(self):
        settings = get_settings()
        self.layer._highlight_width = (
            settings.appearance.highlight.highlight_thickness
        )
        self.layer._highlight_color = (
            settings.appearance.highlight.highlight_color
        )

        # Compute the vertices and faces of any shape outlines
        vertices, faces = self.layer._outline_shapes()

        if vertices is None or len(vertices) == 0 or len(faces) == 0:
            vertices = np.zeros((3, self.layer._slice_input.ndisplay))
            faces = np.array([[0, 1, 2]])

        self.node.shape_highlights.set_data(
            vertices=vertices,
            faces=faces,
            color=self.layer._highlight_color,
        )

        # Compute the location and properties of the vertices and box that
        # need to get rendered
        (
            vertices,
            face_color,
            edge_color,
            pos,
            _,
        ) = self.layer._compute_vertices_and_box()

        width = settings.appearance.highlight.highlight_thickness

        if vertices is None or len(vertices) == 0:
            vertices = np.zeros((1, self.layer._slice_input.ndisplay))
            size = 0
        else:
            size = self.layer._vertex_size

        self.node.highlight_vertices.set_data(
            vertices,
            size=size,
            face_color=face_color,
            edge_color=edge_color,
            edge_width=width,
        )

        if pos is None or len(pos) == 0:
            pos = np.zeros((1, self.layer._slice_input.ndisplay))
            width = 0

        self.node.highlight_lines.set_data(
            pos=pos, color=edge_color, width=width
        )

    def _update_text(self, *, update_node=True):
        """Function to update the text node properties

        Parameters
        ----------
        update_node : bool
            If true, update the node after setting the properties
        """
        update_text(node=self._get_text_node(), layer=self.layer)
        if update_node:
            self.node.update()

    def _get_text_node(self):
        """Function to get the text node from the Compound visual"""
        return self.node.text

    def _on_text_change(self, event=None):
        if event is not None:
            if event.type == 'blending':
                self._on_blending_change(event)
                return
            if event.type == 'values':
                return
        self._update_text()

    def _on_blending_change(self):
        """Function to set the blending mode"""
        shapes_blending_kwargs = BLENDING_MODES[self.layer.blending]
        self.node.set_gl_state(**shapes_blending_kwargs)

        text_node = self._get_text_node()
        text_blending_kwargs = BLENDING_MODES[self.layer.text.blending]
        text_node.set_gl_state(**text_blending_kwargs)
        self.node.update()

    def reset(self):
        super().reset()
        self._on_highlight_change()
        self._on_blending_change()

    def close(self):
        """Vispy visual is closing."""
        disconnect_events(self.layer.text.events, self)
        super().close()



"""Qt 'Help' menu Actions."""

import sys
from functools import partial
from webbrowser import open as web_open

from app_model.types import Action, KeyBindingRule, KeyCode, KeyMod
from packaging.version import parse

from napari import __version__
from napari._app_model.constants import MenuGroup, MenuId
from napari._qt.dialogs.qt_about import QtAbout
from napari._qt.qt_main_window import Window
from napari._qt.widgets.qt_logger import LogWidget
from napari.utils.translations import trans


def _show_about(window: Window):
    QtAbout.showAbout(window._qt_window)


def _show_logs(window: Window):
    window.add_dock_widget(
        LogWidget(), name='logger', area='bottom', tabify=True
    )


v = parse(__version__)
VERSION = 'dev' if v.is_devrelease or v.is_prerelease else str(v.base_version)

HELP_URLS: dict[str, str] = {
    'getting_started': f'https://napari.org/{VERSION}/tutorials/start_index.html',
    'tutorials': f'https://napari.org/{VERSION}/tutorials/index.html',
    'layers_guide': f'https://napari.org/{VERSION}/howtos/layers/index.html',
    'examples_gallery': f'https://napari.org/{VERSION}/gallery.html',
    'plugins': f'https://napari.org/{VERSION}/plugins/index.html',
    'contribute': f'https://napari.org/{VERSION}/developers/index.html',
    'release_notes': f'https://napari.org/{VERSION}/release/release_{VERSION.replace(".", "_")}.html',
    'github_issue': 'https://github.com/napari/napari/issues',
    'homepage': 'https://napari.org',
}

Q_HELP_ACTIONS: list[Action] = [
    Action(
        id='napari.window.help.info',
        title=trans._('‎napari Info'),
        callback=_show_about,
        menus=[{'id': MenuId.MENUBAR_HELP, 'group': MenuGroup.RENDER}],
        status_tip=trans._('About napari'),
        keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.Slash)],
    ),
    Action(
        id='napari.window.help.about_macos',
        title=trans._('About napari'),
        callback=_show_about,
        menus=[
            {
                'id': MenuId.MENUBAR_HELP,
                'group': MenuGroup.RENDER,
                'when': sys.platform == 'darwin',
            }
        ],
        status_tip=trans._('About napari'),
    ),
    Action(
        id='napari.window.help.getting_started',
        title=trans._('Getting started'),
        callback=partial(web_open, url=HELP_URLS['getting_started']),
        menus=[{'id': MenuId.MENUBAR_HELP}],
    ),
    Action(
        id='napari.window.help.tutorials',
        title=trans._('Tutorials'),
        callback=partial(web_open, url=HELP_URLS['tutorials']),
        menus=[{'id': MenuId.MENUBAR_HELP}],
    ),
    Action(
        id='napari.window.help.layers_guide',
        title=trans._('Using Layers Guides'),
        callback=partial(web_open, url=HELP_URLS['layers_guide']),
        menus=[{'id': MenuId.MENUBAR_HELP}],
    ),
    Action(
        id='napari.window.help.examples',
        title=trans._('Examples Gallery'),
        callback=partial(web_open, url=HELP_URLS['examples_gallery']),
        menus=[{'id': MenuId.MENUBAR_HELP}],
    ),
    Action(
        id='napari.window.help.plugins',
        title=trans._('Extend with Plugins'),
        callback=partial(web_open, url=HELP_URLS['plugins']),
        menus=[{'id': MenuId.MENUBAR_HELP}],
    ),
    Action(
        id='napari.window.help.release_notes',
        title=trans._('Release Notes'),
        callback=partial(web_open, url=HELP_URLS['release_notes']),
        menus=[
            {
                'id': MenuId.MENUBAR_HELP,
                'when': VERSION != 'dev',
                'group': MenuGroup.NAVIGATION,
            }
        ],
    ),
    Action(
        id='napari.window.help.github_issue',
        title=trans._('Report an issue on GitHub'),
        callback=partial(web_open, url=HELP_URLS['github_issue']),
        menus=[
            {
                'id': MenuId.MENUBAR_HELP,
                'when': VERSION == 'dev',
                'group': MenuGroup.NAVIGATION,
            }
        ],
    ),
    Action(
        id='napari.window.help.contribute',
        title=trans._('Contribute to napari'),
        callback=partial(web_open, url=HELP_URLS['contribute']),
        menus=[{'id': MenuId.MENUBAR_HELP, 'group': MenuGroup.NAVIGATION}],
    ),
    Action(
        id='napari.window.help.homepage',
        title=trans._('napari homepage'),
        callback=partial(web_open, url=HELP_URLS['homepage']),
        menus=[{'id': MenuId.MENUBAR_HELP, 'group': MenuGroup.NAVIGATION}],
    ),
    Action(
        id='napari.window.help.show_logs',
        title=trans._('Show logs'),
        callback=_show_logs,
        menus=[{'id': MenuId.MENUBAR_HELP, 'group': MenuGroup.RENDER}],
        status_tip=trans._('View and filter logs'),
    ),
]



from napari.components._viewer_constants import CanvasPosition
from napari.layers.base._base_constants import Blending
from napari.utils.events import EventedModel


class Overlay(EventedModel):
    """
    Overlay evented model.

    An overlay is a renderable entity meant to display additional information
    on top of the layer data, but is not data per se.
    For example: a scale bar, a color bar, axes, bounding boxes, etc.

    Attributes
    ----------
    visible : bool
        If the overlay is visible or not.
    opacity : float
        The opacity of the overlay. 0 is fully transparent.
    order : int
        The rendering order of the overlay: lower numbers get rendered first.
    """

    visible: bool = False
    opacity: float = 1
    order: int = 10**6
    blending: Blending

    def __hash__(self):
        return id(self)


class CanvasOverlay(Overlay):
    """
    Canvas overlay model.

    Canvas overlays live in canvas space; they do not live in the 2- or 3-dimensional scene being rendered, but in the 2D space of the screen.
    For example: scale bars, colormap bars, etc.

    Attributes
    ----------
    position : CanvasPosition
        The position of the overlay in the canvas.
    visible : bool
        If the overlay is visible or not.
    opacity : float
        The opacity of the overlay. 0 is fully transparent.
    order : int
        The rendering order of the overlay: lower numbers get rendered first.
    """

    position: CanvasPosition | tuple[int, int] = CanvasPosition.BOTTOM_RIGHT
    blending: Blending = Blending.TRANSLUCENT_NO_DEPTH


class SceneOverlay(Overlay):
    """
    Scene overlay model.

    Scene overlays live in the 2- or 3-dimensional space of the rendered data.
    For example: bounding boxes, data grids, etc.

    Attributes
    ----------
    visible : bool
        If the overlay is visible or not.
    opacity : float
        The opacity of the overlay. 0 is fully transparent.
    order : int
        The rendering order of the overlay: lower numbers get rendered first.
    """

    blending: Blending = Blending.TRANSLUCENT



import numpy as np
from qtpy.QtCore import QSize, Qt, Signal
from qtpy.QtGui import QColor, QIntValidator, QPainter, QPainterPath, QPen
from qtpy.QtWidgets import (
    QFrame,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QSlider,
    QVBoxLayout,
    QWidget,
)

from napari._qt.widgets.qt_color_swatch import QColorSwatchEdit
from napari.utils.translations import translator

trans = translator.load()


class QtStar(QFrame):
    """Creates a star for the preview pane in the highlight widget.

    Parameters
    ----------
    value : int
        The line width of the star.
    """

    def __init__(
        self,
        parent: QWidget = None,
        value: int | None = None,
    ) -> None:
        super().__init__(parent)
        self._value = value
        self._color = QColor(135, 206, 235)

    def sizeHint(self):
        """Override Qt sizeHint."""
        return QSize(100, 100)

    def minimumSizeHint(self):
        """Override Qt minimumSizeHint."""
        return QSize(100, 100)

    def paintEvent(self, e):
        """Paint star on frame."""
        qp = QPainter()
        qp.begin(self)

        self.drawStar(qp)
        qp.end()

    def value(self):
        """Return value of star widget.

        Returns
        -------
        int
            The value of the star widget.

        """
        return self._value

    def setValue(self, value: int, color: QColor = None):
        """Set line width value of star widget.

        Parameters
        ----------
        value : int
            line width value for star
        """

        self._value = value
        if color is not None:
            self._color = color
        self.update()

    def drawStar(self, qp):
        """Draw a star in the preview pane.

        Parameters
        ----------
        qp : QPainter object
        """
        width = self.rect().width()
        height = self.rect().height()
        col = self._color
        pen = QPen(col, self._value)
        pen.setJoinStyle(Qt.PenJoinStyle.MiterJoin)
        qp.setPen(pen)

        path = QPainterPath()

        # draw pentagram
        star_center_x = width / 2
        star_center_y = height / 2
        # make sure the star equal no matter the size of the qframe
        radius_outer = width * 0.35 if width < height else height * 0.35
        # start at the top point of the star and move counter clockwise to draw the path.
        # every other point is the shorter radius (1/(1+golden_ratio)) of the larger radius
        golden_ratio = (1 + np.sqrt(5)) / 2
        radius_inner = radius_outer / (1 + golden_ratio)
        theta_start = np.pi / 2
        theta_inc = (2 * np.pi) / 10
        for n in range(11):
            theta = theta_start + (n * theta_inc)
            theta = np.mod(theta, 2 * np.pi)
            if np.mod(n, 2) == 0:
                # use radius_outer
                x = radius_outer * np.cos(theta)
                y = radius_outer * np.sin(theta)

            else:
                # use radius_inner
                x = radius_inner * np.cos(theta)
                y = radius_inner * np.sin(theta)

            x_adj = star_center_x - x
            y_adj = star_center_y - y + 3
            if n == 0:
                path.moveTo(x_adj, y_adj)
            else:
                path.lineTo(x_adj, y_adj)

        qp.drawPath(path)


class QtTriangle(QFrame):
    """Draw the triangle in highlight widget.
    Parameters
    ----------
    value : int
        Current value of the highlight size.
    min_value : int
        Minimum value possible for highlight size.
    max_value : int
        Maximum value possible for highlight size.
    """

    valueChanged = Signal(int)

    def __init__(
        self,
        parent: QWidget = None,
        value: int = 1,
        min_value: int = 1,
        max_value: int = 10,
    ) -> None:
        super().__init__(parent)
        self._max_value = max_value
        self._min_value = min_value
        self._value = value
        self._color = QColor(135, 206, 235)

    def mousePressEvent(self, event):
        """When mouse is clicked, adjust to new values."""
        # set value based on position of event
        perc = event.pos().x() / self.rect().width()
        value = ((self._max_value - self._min_value) * perc) + self._min_value
        self.setValue(value)

    def paintEvent(self, e):
        """Paint triangle on frame."""
        qp = QPainter()
        qp.begin(self)
        self.drawTriangle(qp)
        perc = (self._value - self._min_value) / (
            self._max_value - self._min_value
        )
        self.drawLine(qp, self.rect().width() * perc)
        qp.end()

    def sizeHint(self):
        """Override Qt sizeHint."""
        return QSize(75, 30)

    def minimumSizeHint(self):
        """Override Qt minimumSizeHint."""
        return QSize(75, 30)

    def drawTriangle(self, qp):
        """Draw triangle.

        Parameters
        ----------
        qp : QPainter object
        """
        width = self.rect().width()

        col = self._color
        qp.setPen(QPen(col, 1))
        qp.setBrush(col)
        path = QPainterPath()

        height = 10
        path.moveTo(0, height)

        path.lineTo(width, height)

        path.lineTo(width, 0)
        path.closeSubpath()

        qp.drawPath(path)

    def value(self):
        """Return value of triangle widget.

        Returns
        -------
        int
            Current value of triangle widget.
        """
        return self._value

    def setValue(self, value, color=None):
        """Set value for triangle widget.

        Parameters
        ----------
        value : int
            Value to use for line in triangle widget.
        """
        self._value = value
        if color is not None:
            self._color = color
        self.update()

    def minimum(self):
        """Return minimum value.

        Returns
        -------
        int
            Mininum value of triangle widget.
        """
        return self._min_value

    def maximum(self):
        """Return maximum value.

        Returns
        -------
        int
            Maximum value of triangle widget.
        """
        return self._max_value

    def setMinimum(self, value: int):
        """Set minimum value

        Parameters
        ----------
        value : int
            Minimum value of triangle.
        """
        self._min_value = value
        self._value = max(self._value, value)

    def setMaximum(self, value: int):
        """Set maximum value.

        Parameters
        ----------
        value : int
            Maximum value of triangle.
        """
        self._max_value = value

        self._value = min(self._value, value)

    def drawLine(self, qp, value: int):
        """Draw line on triangle indicating value.

        Parameters
        ----------
        qp : QPainter object
        value : int
            Value of highlight thickness.
        """
        col = QColor('white')
        qp.setPen(QPen(col, 2))
        qp.setBrush(col)
        path = QPainterPath()
        path.moveTo(value, 15)

        path.lineTo(value, 0)
        path.closeSubpath()

        qp.drawPath(path)
        self.valueChanged.emit(self._value)


class QtHighlightPreviewWidget(QWidget):
    """Creates custom widget to set highlight size.

    Parameters
    ----------
    description : str
        Text to explain and display on widget.
    value : int
        Value of highlight size.
    min_value : int
        Minimum possible value of highlight size.
    max_value : int
        Maximum possible value of highlight size.
    unit : str
        Unit of highlight size.
    """

    valueChanged = Signal(dict)

    def __init__(
        self,
        parent: QWidget = None,
        description: str = '',
        value: dict | None = None,
        min_value: int = 1,
        max_value: int = 10,
        unit: str = 'px',
    ) -> None:
        super().__init__(parent)

        self.setGeometry(300, 300, 125, 110)
        if value is None:
            value = {
                'highlight_thickness': 1,
                'highlight_color': [0.0, 0.6, 1.0, 1.0],
            }
        self._value = value
        self._thickness_value = (
            value['highlight_thickness'] or self.fontMetrics().height()
        )
        self._color_value = value['highlight_color'] or [0.0, 0.6, 1.0, 1.0]
        self._min_value = min_value
        self._max_value = max_value

        # Widget
        self._color_swatch_edit = QColorSwatchEdit(
            self, initial_color=self._color_value
        )
        self._lineedit = QLineEdit()
        self._description = QLabel(self)
        self._unit = QLabel(self)
        self._slider = QSlider(Qt.Orientation.Horizontal)
        self._triangle = QtTriangle(self)
        self._slider_min_label = QLabel(self)
        self._slider_max_label = QLabel(self)
        self._preview = QtStar(self)
        self._preview_label = QLabel(self)
        self._validator = QIntValidator(min_value, max_value, self)

        # Widgets setup
        self._description.setText(description)
        self._description.setWordWrap(True)
        self._unit.setText(unit)
        self._unit.setAlignment(Qt.AlignmentFlag.AlignBottom)
        self._lineedit.setValidator(self._validator)
        self._lineedit.setAlignment(Qt.AlignmentFlag.AlignRight)
        self._lineedit.setAlignment(Qt.AlignmentFlag.AlignBottom)
        self._slider_min_label.setText(str(min_value))
        self._slider_min_label.setAlignment(Qt.AlignmentFlag.AlignBottom)
        self._slider_max_label.setText(str(max_value))
        self._slider_max_label.setAlignment(Qt.AlignmentFlag.AlignBottom)
        self._slider.setMinimum(min_value)
        self._slider.setMaximum(max_value)
        self._preview.setValue(self._thickness_value)
        self._triangle.setValue(self._thickness_value)
        self._triangle.setMinimum(min_value)
        self._triangle.setMaximum(max_value)
        self._preview_label.setText(trans._('Preview'))
        self._preview_label.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        self._preview_label.setAlignment(Qt.AlignmentFlag.AlignBottom)
        self._preview.setStyleSheet('border: 1px solid white;')

        # Signals
        self._slider.valueChanged.connect(self._update_thickness_value)
        self._lineedit.textChanged.connect(self._update_thickness_value)
        self._triangle.valueChanged.connect(self._update_thickness_value)
        self._color_swatch_edit.color_changed.connect(self._update_color_value)

        # Layout
        triangle_layout = QHBoxLayout()
        triangle_layout.addWidget(self._triangle)
        triangle_layout.setContentsMargins(6, 35, 6, 0)
        triangle_slider_layout = QVBoxLayout()
        triangle_slider_layout.addLayout(triangle_layout)
        triangle_slider_layout.setContentsMargins(0, 0, 0, 0)
        triangle_slider_layout.addWidget(self._slider)
        triangle_slider_layout.setAlignment(Qt.AlignmentFlag.AlignVCenter)

        # Bottom row layout
        lineedit_layout = QVBoxLayout()
        lineedit_layout.addWidget(self._color_swatch_edit)
        lineedit_layout.addWidget(self._lineedit)
        lineedit_layout.setAlignment(Qt.AlignmentFlag.AlignBottom)
        bottom_left_layout = QHBoxLayout()
        bottom_left_layout.addLayout(lineedit_layout)
        bottom_left_layout.addWidget(self._unit)
        bottom_left_layout.addWidget(self._slider_min_label)
        bottom_left_layout.addLayout(triangle_slider_layout)
        bottom_left_layout.addWidget(self._slider_max_label)
        bottom_left_layout.setAlignment(Qt.AlignmentFlag.AlignBottom)

        left_layout = QVBoxLayout()
        left_layout.addWidget(self._description)
        left_layout.addLayout(bottom_left_layout)
        left_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)

        preview_label_layout = QHBoxLayout()
        preview_label_layout.addWidget(self._preview_label)
        preview_label_layout.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        preview_layout = QVBoxLayout()
        preview_layout.addWidget(self._preview)
        preview_layout.addLayout(preview_label_layout)
        preview_layout.setAlignment(Qt.AlignmentFlag.AlignCenter)

        layout = QHBoxLayout()
        layout.addLayout(left_layout)
        layout.addLayout(preview_layout)

        self.setLayout(layout)

        self._refresh()

    def _update_thickness_value(self, thickness_value):
        """Update highlight thickness value.

        Parameters
        ----------
        thickness_value : int
            Highlight thickness value.
        """
        if thickness_value == '':
            return
        thickness_value = int(thickness_value)
        thickness_value = max(
            min(thickness_value, self._max_value), self._min_value
        )
        if thickness_value == self._thickness_value:
            return
        self._thickness_value = thickness_value
        self._value['highlight_thickness'] = self._thickness_value
        self.valueChanged.emit(self._value)
        self._refresh()

    def _update_color_value(self, color_value):
        """Update highlight color value.

        Parameters
        ----------
        color_value : List[float]
            Highlight color value as a list of floats.
        """
        if isinstance(color_value, np.ndarray):
            color_value = color_value.tolist()
        if color_value == self._color_value:
            return
        if color_value == '':
            return
        self._color_value = color_value
        self._value['highlight_color'] = self._color_value
        self.valueChanged.emit(self._value)
        self._refresh()

    def _refresh(self):
        """Set every widget value to the new set value."""
        self.blockSignals(True)
        # Transform color value from a float representation
        # (values from 0.0 to 1.0) to RGB values (values from 0 to 255)
        # to set widgets color.
        color = QColor(*[int(np.ceil(v * 255)) for v in self._color_value[:3]])
        self._lineedit.setText(str(self._thickness_value))
        self._slider.setValue(self._thickness_value)
        self._triangle.setValue(self._thickness_value, color=color)
        self._preview.setValue(self._thickness_value, color=color)
        self._color_swatch_edit.setColor(self._color_value)
        self.blockSignals(False)

    def value(self):
        """Return current value.

        Returns
        -------
        dict
            Current value of highlight widget (thickness and color).
        """
        return self._value

    def setValue(self, value: dict):
        """Set new value and update widget.

        Parameters
        ----------
        value : dict
            Highlight value (thickness and color).
        """
        self._update_thickness_value(value.get('highlight_thickness', ''))
        self._update_color_value(value.get('highlight_color', ''))
        self._refresh()

    def description(self):
        """Return the description text.

        Returns
        -------
        str
            Current text in description.
        """
        return self._description.text()

    def setDescription(self, text):
        """Set the description text.

        Parameters
        ----------
        text : str
            Text to use in description box.
        """
        self._description.setText(text)

    def unit(self):
        """Return highlight value unit text.

        Returns
        -------
        str
            Current text in unit text.
        """
        return self._unit.text()

    def setUnit(self, text):
        """Set highlight value unit.

        Parameters
        ----------
        text : str
            Text used to describe units.
        """
        self._unit.setText(text)

    def setMinimum(self, value):
        """Set minimum highlight value for star, triangle, text and slider.

        Parameters
        ----------
        value : int
            Minimum highlight value.
        """
        value = int(value)
        if value >= self._max_value:
            raise ValueError(
                trans._(
                    'Minimum value must be smaller than {max_value}',
                    deferred=True,
                    max_value=self._max_value,
                )
            )
        self._min_value = value
        self._slider_min_label.setText(str(value))
        self._slider.setMinimum(value)
        self._triangle.setMinimum(value)
        self._thickness_value = max(
            self._value['highlight_thickness'], self._min_value
        )
        self._value['highlight_thickness'] = self._thickness_value
        self._refresh()

    def minimum(self):
        """Return minimum highlight value.

        Returns
        -------
        int
            Minimum value of highlight widget.
        """
        return self._min_value

    def setMaximum(self, value):
        """Set maximum highlight value.

        Parameters
        ----------
        value : int
            Maximum highlight value.
        """
        value = int(value)
        if value <= self._min_value:
            raise ValueError(
                trans._(
                    'Maximum value must be larger than {min_value}',
                    deferred=True,
                    min_value=self._min_value,
                )
            )
        self._max_value = value
        self._slider_max_label.setText(str(value))
        self._slider.setMaximum(value)
        self._triangle.setMaximum(value)
        self._thickness_value = min(
            self._value['highlight_thickness'], self._max_value
        )
        self._value['highlight_thickness'] = self._thickness_value
        self._refresh()

    def maximum(self):
        """Return maximum highlight value.

        Returns
        -------
        int
            Maximum value of highlight widget.
        """
        return self._max_value



from typing import Literal

import numpy as np

__all__ = (
    'CoordinateArray',
    'CoordinateArray2D',
    'CoordinateArray3D',
    'EdgeArray',
    'TriangleArray',
)

CoordinateArray2D = np.ndarray[tuple[int, Literal[2]], np.dtype[np.float32]]
CoordinateArray3D = np.ndarray[tuple[int, Literal[3]], np.dtype[np.float32]]
CoordinateArray = CoordinateArray2D | CoordinateArray3D
EdgeArray = np.ndarray[tuple[int, Literal[2]], np.dtype[np.int64]]
TriangleArray = np.ndarray[tuple[int, Literal[3]], np.dtype[np.int32]]



from __future__ import annotations

from vispy.scene.visuals import Compound, Line, Text

from napari._vispy.visuals.clipping_planes_mixin import ClippingPlanesMixin
from napari._vispy.visuals.markers import Markers


class PointsVisual(ClippingPlanesMixin, Compound):
    """
    Compound vispy visual for point visualization with
    clipping planes functionality

    Components:
        - Markers for points (vispy.MarkersVisual)
        - Markers for selection highlights (vispy.MarkersVisual)
        - Lines for highlights (vispy.LineVisual)
        - Text labels (vispy.TextVisual)
    """

    def __init__(self) -> None:
        super().__init__(
            [
                Markers(),
                Markers(),
                Line(),
                Text(),
            ]
        )
        self.scaling = True

    @property
    def points_markers(self) -> Markers:
        """Points markers visual"""
        return self._subvisuals[0]

    @property
    def selection_markers(self) -> Markers:
        """Highlight markers visual"""
        return self._subvisuals[1]

    @property
    def highlight_lines(self) -> Line:
        """Highlight lines visual"""
        return self._subvisuals[2]

    @property
    def text(self) -> Text:
        """Text labels visual"""
        return self._subvisuals[3]

    @property
    def scaling(self) -> bool:
        """
        Scaling property for both the markers visuals. If set to true,
        the points rescale based on zoom (i.e: constant world-space size)
        """
        return self.points_markers.scaling == 'visual'

    @scaling.setter
    def scaling(self, value: bool) -> None:
        scaling_txt = 'visual' if value else 'fixed'
        self.points_markers.scaling = scaling_txt
        self.selection_markers.scaling = scaling_txt

    @property
    def antialias(self) -> float:
        return self.points_markers.antialias

    @antialias.setter
    def antialias(self, value: float) -> None:
        self.points_markers.antialias = value
        self.selection_markers.antialias = value

    @property
    def spherical(self) -> bool:
        return self.points_markers.spherical

    @spherical.setter
    def spherical(self, value: bool) -> None:
        self.points_markers.spherical = value

    @property
    def canvas_size_limits(self) -> tuple[int, int]:
        return self.points_markers.canvas_size_limits

    @canvas_size_limits.setter
    def canvas_size_limits(self, value: tuple[int, int]) -> None:
        self.points_markers.canvas_size_limits = value
        self.selection_markers.canvas_size_limits = value



from qtpy.QtCore import Qt

from napari._app_model import get_app_model
from napari._qt.widgets.qt_command_palette import QCommandPalette


def test_move_command_palette(qtbot):
    qt_command_palette = QCommandPalette()
    qtbot.addWidget(qt_command_palette)
    qt_command_palette._line.setText('napari')
    assert qt_command_palette._list._selected_index == 0
    qtbot.keyClick(qt_command_palette._line, Qt.Key.Key_Down)
    assert qt_command_palette._list._selected_index == 1
    qtbot.keyClick(qt_command_palette._line, Qt.Key.Key_Up)
    assert qt_command_palette._list._selected_index == 0
    qtbot.keyClick(qt_command_palette._line, Qt.Key.Key_PageDown)
    assert qt_command_palette._list._selected_index > 3
    qtbot.keyClick(qt_command_palette._line, Qt.Key.Key_PageUp)
    assert qt_command_palette._list._selected_index == 0


def _no_op():
    pass


def test_command_palette_app_model_sync(qtbot, mock_app_model):
    qt_command_palette = QCommandPalette()
    qtbot.addWidget(qt_command_palette)

    n_commands = len(qt_command_palette._list.all_commands)
    app = get_app_model()
    deregister_noop = app.register_action(
        'napari.noop', 'no-op', callback=_no_op
    )

    assert len(qt_command_palette._list.all_commands) == n_commands + 1

    deregister_noop()

    assert len(qt_command_palette._list.all_commands) == n_commands



from typing import TypeVar

T = TypeVar('T')


def _coerce_extensions_to_globs(reader_settings: dict[str, T]) -> dict[str, T]:
    """Coerce existing reader settings for file extensions to glob patterns"""
    new_settings = {}
    for pattern, reader in reader_settings.items():
        if pattern.startswith('.') and '*' not in pattern:
            pattern = f'*{pattern}'
        new_settings[pattern] = reader
    return new_settings



import time
import weakref
from concurrent.futures import Future, wait
from dataclasses import dataclass
from threading import RLock, current_thread, main_thread
from typing import Any

import numpy as np
import pytest

from napari._tests.utils import DEFAULT_TIMEOUT_SECS, LockableData
from napari.components import Dims
from napari.components._layer_slicer import _LayerSlicer
from napari.layers import Image, Labels, Points

# The following fakes are used to control execution of slicing across
# multiple threads, while also allowing us to mimic real classes
# (like layers) in the code base. This allows us to assert state and
# conditions that may only be temporarily true at different stages of
# an asynchronous task.


@dataclass(frozen=True)
class FakeSliceResponse:
    id: int


@dataclass(frozen=True)
class FakeSliceRequest:
    id: int
    lock: RLock

    def __call__(self) -> FakeSliceResponse:
        assert current_thread() != main_thread()
        with self.lock:
            return FakeSliceResponse(id=self.id)


class FakeAsyncLayer:
    def __init__(self) -> None:
        self._last_slice_id: int = 0
        self._slice_request_count: int = 0
        self.slice_count: int = 0
        self.visible: bool = True
        self.lock: RLock = RLock()

    def _make_slice_request(self, dims: Dims) -> FakeSliceRequest:
        assert current_thread() == main_thread()
        self._slice_request_count += 1
        return FakeSliceRequest(id=self._slice_request_count, lock=self.lock)

    def _update_slice_response(self, response: FakeSliceResponse):
        self.slice_count = response.id

    def _slice_dims(self, *args, **kwargs) -> None:
        self.slice_count += 1

    def _set_unloaded_slice_id(self, slice_id: int) -> None:
        self._last_slice_id = slice_id


class FakeSyncLayer:
    def __init__(self) -> None:
        self.slice_count: int = 0
        self.visible: bool = True

    def _slice_dims(self, *args, **kwargs) -> None:
        self.slice_count += 1


@pytest.fixture
def layer_slicer():
    layer_slicer = _LayerSlicer()
    layer_slicer._force_sync = False
    yield layer_slicer
    layer_slicer.shutdown()


def test_submit_with_one_async_layer_no_block(layer_slicer):
    layer = FakeAsyncLayer()

    future = layer_slicer.submit(layers=[layer], dims=Dims())

    assert _wait_for_response(future)[layer].id == 1
    assert _wait_for_response(future)[layer].id == 1


def test_submit_with_multiple_async_layer_no_block(layer_slicer):
    layer1 = FakeAsyncLayer()
    layer2 = FakeAsyncLayer()

    future = layer_slicer.submit(layers=[layer1, layer2], dims=Dims())

    assert _wait_for_response(future)[layer1].id == 1
    assert _wait_for_response(future)[layer2].id == 1


def test_submit_emits_ready_event_when_done(layer_slicer):
    layer = FakeAsyncLayer()
    event_result = None

    def on_done(event):
        nonlocal event_result
        event_result = event.value

    layer_slicer.events.ready.connect(on_done)

    future = layer_slicer.submit(layers=[layer], dims=Dims())
    actual_result = _wait_for_result(future)

    assert actual_result is event_result


def test_submit_with_one_sync_layer(layer_slicer):
    layer = FakeSyncLayer()
    assert layer.slice_count == 0

    future = layer_slicer.submit(layers=[layer], dims=Dims())

    assert layer.slice_count == 1
    assert future is None


def test_submit_with_multiple_sync_layer(layer_slicer):
    layer1 = FakeSyncLayer()
    layer2 = FakeSyncLayer()
    assert layer1.slice_count == 0
    assert layer2.slice_count == 0

    future = layer_slicer.submit(layers=[layer1, layer2], dims=Dims())

    assert layer1.slice_count == 1
    assert layer2.slice_count == 1
    assert future is None


def test_submit_with_mixed_layers(layer_slicer):
    layer1 = FakeAsyncLayer()
    layer2 = FakeSyncLayer()
    assert layer1.slice_count == 0
    assert layer2.slice_count == 0

    future = layer_slicer.submit(layers=[layer1, layer2], dims=Dims())

    assert layer2.slice_count == 1
    assert _wait_for_response(future)[layer1].id == 1
    assert layer2 not in _wait_for_response(future)


def test_submit_lock_blocking(layer_slicer):
    dims = Dims()
    layer = FakeAsyncLayer()

    assert layer.slice_count == 0
    with layer.lock:
        blocked = layer_slicer.submit(layers=[layer], dims=dims)
        assert not blocked.done()

    assert _wait_for_response(blocked)[layer].id == 1


def test_submit_multiple_calls_cancels_pending(layer_slicer):
    dims = Dims()
    layer = FakeAsyncLayer()

    with layer.lock:
        blocked = layer_slicer.submit(layers=[layer], dims=dims)
        _wait_until_running(blocked)
        pending = layer_slicer.submit(layers=[layer], dims=dims)
        assert not pending.running()
        layer_slicer.submit(layers=[layer], dims=dims)
        assert not blocked.done()

    assert pending.cancelled()


def test_submit_mixed_allows_sync_to_run(layer_slicer):
    """ensure that a blocked async slice doesn't block sync slicing"""
    dims = Dims()
    layer1 = FakeAsyncLayer()
    layer2 = FakeSyncLayer()
    with layer1.lock:
        blocked = layer_slicer.submit(layers=[layer1], dims=dims)
        layer_slicer.submit(layers=[layer2], dims=dims)
        assert layer2.slice_count == 1
        assert not blocked.done()

    assert _wait_for_response(blocked)[layer1].id == 1


def test_submit_mixed_allows_sync_to_run_one_slicer_call(layer_slicer):
    """ensure that a blocked async slice doesn't block sync slicing"""
    dims = Dims()
    layer1 = FakeAsyncLayer()
    layer2 = FakeSyncLayer()
    with layer1.lock:
        blocked = layer_slicer.submit(layers=[layer1, layer2], dims=dims)

        assert layer2.slice_count == 1
        assert not blocked.done()

    assert _wait_for_response(blocked)[layer1].id == 1


def test_submit_with_multiple_async_layer_with_all_locked(
    layer_slicer,
):
    """ensure that if only all layers are locked, none continue"""
    dims = Dims()
    layer1 = FakeAsyncLayer()
    layer2 = FakeAsyncLayer()

    with layer1.lock, layer2.lock:
        blocked = layer_slicer.submit(layers=[layer1, layer2], dims=dims)
        assert not blocked.done()

    assert _wait_for_response(blocked)[layer1].id == 1
    assert _wait_for_response(blocked)[layer2].id == 1


def test_submit_task_to_layers_lock(layer_slicer):
    """ensure that if only one layer has a lock, the non-locked layer
    can continue"""
    dims = Dims()
    layer = FakeAsyncLayer()

    with layer.lock:
        task = layer_slicer.submit(layers=[layer], dims=dims)
        assert task in layer_slicer._layers_to_task.values()

    assert _wait_for_response(task)[layer].id == 1
    assert task not in layer_slicer._layers_to_task


def test_submit_exception_main_thread(layer_slicer):
    """Exception is raised on the main thread from an error on the main
    thread immediately when the task is created."""

    class FakeAsyncLayerError(FakeAsyncLayer):
        def _make_slice_request(self, dims) -> FakeSliceRequest:
            raise RuntimeError('_make_slice_request')

    layer = FakeAsyncLayerError()
    with pytest.raises(RuntimeError, match='_make_slice_request'):
        layer_slicer.submit(layers=[layer], dims=Dims())


def test_submit_exception_subthread_on_result(layer_slicer):
    """Exception is raised on the main thread from an error on a subthread
    only after result is called, not upon submission of the task."""

    @dataclass(frozen=True)
    class FakeSliceRequestError(FakeSliceRequest):
        def __call__(self) -> FakeSliceResponse:
            assert current_thread() != main_thread()
            raise RuntimeError('FakeSliceRequestError')

    class FakeAsyncLayerError(FakeAsyncLayer):
        def _make_slice_request(self, dims: Dims) -> FakeSliceRequestError:
            self._slice_request_count += 1
            return FakeSliceRequestError(
                id=self._slice_request_count, lock=self.lock
            )

    layer = FakeAsyncLayerError()
    future = layer_slicer.submit(layers=[layer], dims=Dims())

    done, _ = wait([future], timeout=DEFAULT_TIMEOUT_SECS)
    assert done, 'Test future did not complete within timeout.'
    with pytest.raises(RuntimeError, match='FakeSliceRequestError'):
        _wait_for_response(future)


def test_wait_until_idle(layer_slicer, single_threaded_executor):
    dims = Dims()
    layer = FakeAsyncLayer()

    with layer.lock:
        slice_future = layer_slicer.submit(layers=[layer], dims=dims)
        _wait_until_running(slice_future)
        # The slice task has started, but has not finished yet
        # because we are holding the layer's slicing lock.
        assert len(layer_slicer._layers_to_task) > 0
        # We can't call wait_until_idle on this thread because we're
        # holding the layer's slice lock, so submit it to be executed
        # on another thread and also wait for it to start.
        wait_future = single_threaded_executor.submit(
            layer_slicer.wait_until_idle,
            timeout=DEFAULT_TIMEOUT_SECS,
        )
        _wait_until_running(wait_future)

    _wait_for_result(wait_future)
    assert len(layer_slicer._layers_to_task) == 0


def test_force_sync_on_sync_layer(layer_slicer):
    layer = FakeSyncLayer()

    with layer_slicer.force_sync():
        assert layer_slicer._force_sync
        future = layer_slicer.submit(layers=[layer], dims=Dims())

    assert layer.slice_count == 1
    assert future is None
    assert not layer_slicer._force_sync


def test_force_sync_on_async_layer(layer_slicer):
    layer = FakeAsyncLayer()

    with layer_slicer.force_sync():
        assert layer_slicer._force_sync
        future = layer_slicer.submit(layers=[layer], dims=Dims())

    assert layer.slice_count == 1
    assert future is None


def test_submit_with_one_3d_image(layer_slicer):
    np.random.seed(0)
    data = np.random.rand(8, 7, 6)
    lockable_data = LockableData(data)
    layer = Image(data=lockable_data, multiscale=False)
    dims = Dims(
        ndim=3,
        ndisplay=2,
        range=((0, 8, 1), (0, 7, 1), (0, 6, 1)),
        point=(2, 0, 0),
    )

    with lockable_data.lock:
        future = layer_slicer.submit(layers=[layer], dims=dims)
        assert not future.done()
    layer_result = _wait_for_response(future)[layer]
    np.testing.assert_equal(layer_result.image.view, data[2, :, :])


def test_submit_with_3d_labels(layer_slicer):
    np.random.seed(0)
    data = np.random.randint(20, size=(8, 7, 6))
    lockable_data = LockableData(data)
    layer = Labels(lockable_data, multiscale=False)
    dims = Dims(
        ndim=3,
        ndisplay=2,
        range=((0, 8, 1), (0, 7, 1), (0, 6, 1)),
        point=(2, 0, 0),
    )

    with lockable_data.lock:
        future = layer_slicer.submit(layers=[layer], dims=dims)
        assert not future.done()

    layer_result = _wait_for_response(future)[layer]
    np.testing.assert_equal(layer_result.image.view, data[2, :, :])


def test_submit_with_one_3d_points(layer_slicer):
    """ensure that async slicing of points does not block"""
    np.random.seed(0)
    num_points = 100
    data = np.rint(2.0 * np.random.rand(num_points, 3))
    layer = Points(data=data)

    # Note: We are directly accessing and locking the _data of layer. This
    #       forces a block to ensure that the async slicing call returns
    #       before slicing is complete.
    lockable_internal_data = LockableData(layer._data)
    layer._data = lockable_internal_data
    dims = Dims(
        ndim=3,
        ndisplay=2,
        range=((0, 3, 1), (0, 3, 1), (0, 3, 1)),
        point=(1, 0, 0),
    )

    with lockable_internal_data.lock:
        future = layer_slicer.submit(layers=[layer], dims=dims)
        assert not future.done()


def test_submit_after_shutdown_raises():
    layer_slicer = _LayerSlicer()
    layer_slicer._force_sync = False
    layer_slicer.shutdown()
    with pytest.raises(RuntimeError):
        layer_slicer.submit(layers=[FakeAsyncLayer()], dims=Dims())


def _wait_until_running(future: Future):
    """Waits until the given future is running using a default finite timeout."""
    sleep_secs = 0.01
    total_sleep_secs = 0
    while not future.running():
        time.sleep(sleep_secs)
        total_sleep_secs += sleep_secs
        if total_sleep_secs > DEFAULT_TIMEOUT_SECS:
            raise TimeoutError(
                f'Future did not start running after a timeout of {DEFAULT_TIMEOUT_SECS} seconds.'
            )


# if remove quotes once we are Python 3.9+
def _wait_for_result(future: 'Future[Any]') -> Any:
    """Waits until the given future is finished returns its result."""
    return future.result(timeout=DEFAULT_TIMEOUT_SECS)


# remove quotes in types once we are python 3.9+ only.
def _wait_for_response(
    task: 'Future[dict[weakref.ReferenceType[Any], Any]]',
) -> dict:
    """Waits until the given slice task is finished and returns its result."""
    weak_result = _wait_for_result(task)
    result = {}
    for weak_layer, response in weak_result.items():
        if layer := weak_layer():
            result[layer] = response
    return result



from napari._qt._qplugins._qnpe2 import (
    _rebuild_npe1_plugins_menu,
    _rebuild_npe1_samples_menu,
    _register_qt_actions,
)

__all__ = [
    '_rebuild_npe1_plugins_menu',
    '_rebuild_npe1_samples_menu',
    '_register_qt_actions',
]



from unittest.mock import Mock, patch

import pytest
from qtpy.QtCore import QMutex, QThread, QTimer
from superqt.utils import qdebounced


class _TestThread(QThread):
    def __init__(self) -> None:
        super().__init__()
        self.mutex = QMutex()

    def run(self):
        self.mutex.lock()


@pytest.mark.disable_qthread_start
def test_disable_qthread(qapp):
    t = _TestThread()
    t.mutex.lock()
    t.start()
    assert not t.isRunning()
    t.mutex.unlock()


def test_qthread_running(qtbot):
    t = _TestThread()
    t.mutex.lock()
    t.start()
    assert t.isRunning()
    t.mutex.unlock()
    qtbot.waitUntil(t.isFinished, timeout=2000)


@pytest.mark.disable_qtimer_start
def test_disable_qtimer(qtbot):
    t = QTimer()
    t.setInterval(100)
    t.start()
    assert not t.isActive()

    # As qtbot uses a QTimer in waitUntil, we also test if timer disable does not break it
    th = _TestThread()
    th.mutex.lock()
    th.start()
    assert th.isRunning()
    th.mutex.unlock()
    qtbot.waitUntil(th.isFinished, timeout=2000)
    assert not th.isRunning()


@pytest.mark.usefixtures('_disable_throttling')
@patch('qtpy.QtCore.QTimer.start')
def test_disable_throttle(start_mock):
    mock = Mock()

    @qdebounced(timeout=50)
    def f() -> str:
        mock()

    f()
    start_mock.assert_not_called()
    mock.assert_called_once()


def test_lack_disable_throttle(monkeypatch):
    """This is test showing that if we do not use disable_throttling then timer is started"""
    mock = Mock()
    start_mock = Mock()
    active_mock = Mock(return_value=True)
    monkeypatch.setattr('qtpy.QtCore.QTimer.start', start_mock)
    monkeypatch.setattr('qtpy.QtCore.QTimer.isActive', active_mock)

    @qdebounced(timeout=50)
    def f() -> str:
        mock()

    f()
    start_mock.assert_called_once()
    mock.assert_not_called()



from napari.errors.reader_errors import (
    MultipleReaderError,
    NoAvailableReaderError,
    ReaderPluginError,
)

__all__ = [
    'MultipleReaderError',
    'NoAvailableReaderError',
    'ReaderPluginError',
]



from collections.abc import Iterable
from typing import Any

from qtpy.QtCore import QModelIndex, Qt
from typing_extensions import Self

from napari._qt.containers.qt_list_model import QtListModel
from napari.components import Dims
from napari.utils.events import SelectableEventedList


class AxisModel:
    """View of an axis within a dims model.

    The model keeps track of axis names and allows read / write
    access on the corresponding rollable state of a Dims object.

    Parameters
    ----------
    dims : napari.components.dims.Dims
        Parent Dims object.
    axis : int
        Axis index.

    Attributes
    ----------
    dims : napari.components.dims.Dims
        Dimensions object modeling slicing and displaying.
    axis : int
        Axis index.
    """

    def __init__(self, dims: Dims, axis: int) -> None:
        self.dims = dims
        self.axis = axis

    def __hash__(self) -> int:
        return id(self)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return self.dims.axis_labels[self.axis]

    def __eq__(self, other: object) -> bool:
        # to allow comparisons between a list of AxisModels and the current dims order
        # we need to overload the int and str equality check, this is necessary as the
        # comparison will either be against a list of ints (Dims.order) or a list of
        # strings (Dims.axis_labels)
        if isinstance(other, int):
            return self.axis == other
        if isinstance(other, str):
            return repr(self) == other
        if isinstance(other, AxisModel):
            return (self.dims is other.dims) and (self.axis == other.axis)
        return NotImplemented

    @property
    def rollable(self) -> bool:
        """
        If the axis should be rollable.
        """
        return self.dims.rollable[self.axis]

    @rollable.setter
    def rollable(self, value: bool) -> None:
        rollable = list(self.dims.rollable)
        rollable[self.axis] = value
        self.dims.rollable = tuple(rollable)


class AxisList(SelectableEventedList[AxisModel]):
    def __init__(self, axes: Iterable[AxisModel]):
        super().__init__(axes)

    @classmethod
    def from_dims(cls, dims: Dims) -> Self:
        """Create AxisList instance from Dims object.

        The AxisList is filled with a number of AxisModels based
        on the number of dimensions in the Dims object.

        Parameters
        ----------
        dims : napari.components.dims.Dims
            Dims object to be used for creation.

        Returns
        -------
        AxisList
            A selectable evented list of the viewer axes.
        """
        return cls(AxisModel(dims, d) for d in dims.order)


class QtAxisListModel(QtListModel[AxisModel]):
    def data(self, index: QModelIndex, role: Qt.ItemDataRole):
        if not index.isValid():
            return None
        axis = self.getItem(index)
        if role == Qt.ItemDataRole.DisplayRole:
            return str(axis)
        if role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignCenter
        if role == Qt.ItemDataRole.CheckStateRole:
            return (
                Qt.CheckState.Checked
                if axis.rollable
                else Qt.CheckState.Unchecked
            )
        return super().data(index, role)

    def setData(
        self,
        index: QModelIndex,
        value: Any,
        role: int = Qt.ItemDataRole.EditRole,
    ) -> bool:
        axis = self.getItem(index)
        if role == Qt.ItemDataRole.CheckStateRole:
            axis.rollable = Qt.CheckState(value) == Qt.CheckState.Checked
        elif role == Qt.ItemDataRole.EditRole:
            axis_labels = list(axis.dims.axis_labels)
            axis_labels[axis.axis] = value
            axis.dims.axis_labels = tuple(axis_labels)
        else:
            return super().setData(index, value, role=role)
        self.dataChanged.emit(index, index, [role])
        return True

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        """Returns the item flags for the given `index`.

        This describes the properties of a given item in the model.  We set
        them to be editable, checkable, draggable, droppable, etc...
        If index is not a list, we additionally set `Qt.ItemNeverHasChildren`
        (for optimization). Editable models must return a value containing
        `Qt.ItemIsEditable`.

        See Qt.ItemFlags https://doc.qt.io/qt-5/qt.html#ItemFlag-enum

        Parameters
        ----------
        index : qtpy.QtCore.QModelIndex
            Index to return flags for.

        Returns
        -------
        qtpy.QtCore.Qt.ItemFlags
            ItemFlags specific to the given index.
        """
        if not index.isValid():
            # We only allow drops outside and in between the items
            # (and not inside them), in which case the index is not valid.
            return Qt.ItemFlag.ItemIsDropEnabled

        flags = (
            Qt.ItemFlag.ItemIsSelectable
            | Qt.ItemFlag.ItemIsEditable
            | Qt.ItemFlag.ItemIsUserCheckable
            | Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemNeverHasChildren
            | Qt.ItemFlag.ItemIsDragEnabled
        )
        return flags



# from napari.layers.base.base import Layer
# from napari.utils.events import Event
# from napari.utils.colormaps import AVAILABLE_COLORMAPS

from typing import Any
from warnings import warn

import numpy as np
import pandas as pd

from napari.layers.base import Layer
from napari.layers.tracks._track_utils import TrackManager
from napari.utils.colormaps import AVAILABLE_COLORMAPS, Colormap
from napari.utils.events import Event
from napari.utils.translations import trans


class Tracks(Layer):
    """Tracks layer.

    Parameters
    ----------
    data : array (N, D+1)
        Coordinates for N points in D+1 dimensions. ID,T,(Z),Y,X. The first
        axis is the integer ID of the track. D is either 3 or 4 for planar
        or volumetric timeseries respectively.
    affine : n-D array or napari.utils.transforms.Affine
        (N+1, N+1) affine transformation matrix in homogeneous coordinates.
        The first (N, N) entries correspond to a linear transform and
        the final column is a length N translation vector and a 1 or a napari
        `Affine` transform object. Applied as an extra transform on top of the
        provided scale, rotate, and shear values.
    axis_labels : tuple of str, optional
        Dimension names of the layer data.
        If not provided, axis_labels will be set to (..., 'axis -2', 'axis -1').
    blending : str
        One of a list of preset blending modes that determines how RGB and
        alpha values of the layer visual get mixed. Allowed values are
        {'opaque', 'translucent', and 'additive'}.
    cache : bool
        Whether slices of out-of-core datasets should be cached upon retrieval.
        Currently, this only applies to dask arrays.
    color_by : str
        Track property (from property keys) by which to color vertices.
    colormap : str
        Default colormap to use to set vertex colors. Specialized colormaps,
        relating to specified properties can be passed to the layer via
        colormaps_dict.
    colormaps_dict : dict {str: napari.utils.Colormap}
        Optional dictionary mapping each property to a colormap for that
        property. This allows each property to be assigned a specific colormap,
        rather than having a global colormap for everything.
    experimental_clipping_planes : list of dicts, list of ClippingPlane, or ClippingPlaneList
        Each dict defines a clipping plane in 3D in data coordinates.
        Valid dictionary keys are {'position', 'normal', and 'enabled'}.
        Values on the negative side of the normal are discarded if the plane is enabled.
    features : Dataframe-like
        Features table where each row corresponds to a point and each column
        is a feature.
    graph : dict {int: list}
        Graph representing associations between tracks. Dictionary defines the
        mapping between a track ID and the parents of the track. This can be
        one (the track has one parent, and the parent has >=1 child) in the
        case of track splitting, or more than one (the track has multiple
        parents, but only one child) in the case of track merging.
        See examples/tracks_3d_with_graph.py
    head_length : float
        Length of the positive (forward in time) tails in units of time.
    metadata : dict
        Layer metadata.
    name : str
        Name of the layer.
    opacity : float
        Opacity of the layer visual, between 0.0 and 1.0.
    projection_mode : str
        How data outside the viewed dimensions but inside the thick Dims slice will
        be projected onto the viewed dimenions.
    properties : dict {str: array (N,)}, DataFrame
        Properties for each point. Each property should be an array of length N,
        where N is the number of points.
    rotate : float, 3-tuple of float, or n-D array.
        If a float convert into a 2D rotation matrix using that value as an
        angle. If 3-tuple convert into a 3D rotation matrix, using a yaw,
        pitch, roll convention. Otherwise assume an nD rotation. Angles are
        assumed to be in degrees. They can be converted from radians with
        np.degrees if needed.
    scale : tuple of float
        Scale factors for the layer.
    shear : 1-D array or n-D array
        Either a vector of upper triangular values, or an nD shear matrix with
        ones along the main diagonal.
    tail_length : float
        Length of the positive (backward in time) tails in units of time.
    tail_width : float
        Width of the track tails in pixels.
    translate : tuple of float
        Translation values for the layer.
    units : tuple of str or pint.Unit, optional
        Units of the layer data in world coordinates.
        If not provided, the default units are assumed to be pixels.
    visible : bool
        Whether the layer visual is currently being displayed.
    """

    # The max number of tracks that will ever be used to render the thumbnail
    # If more tracks are present then they are randomly subsampled
    _max_tracks_thumbnail = 1024

    def __init__(
        self,
        data,
        *,
        affine=None,
        axis_labels=None,
        blending='additive',
        cache=True,
        color_by='track_id',
        colormap='turbo',
        colormaps_dict=None,
        experimental_clipping_planes=None,
        features=None,
        graph=None,
        head_length: int = 0,
        metadata=None,
        name=None,
        opacity=1.0,
        projection_mode='none',
        properties=None,
        rotate=None,
        scale=None,
        shear=None,
        tail_length: int = 30,
        tail_width: int = 2,
        translate=None,
        units=None,
        visible=True,
    ) -> None:
        # if not provided with any data, set up an empty layer in 2D+t
        # otherwise convert the data to an np.ndarray
        data = np.empty((0, 4)) if data is None else np.asarray(data)

        # set the track data dimensions (remove ID from data)
        ndim = data.shape[1] - 1

        super().__init__(
            data,
            ndim,
            affine=affine,
            axis_labels=axis_labels,
            blending=blending,
            cache=cache,
            experimental_clipping_planes=experimental_clipping_planes,
            name=name,
            metadata=metadata,
            opacity=opacity,
            projection_mode=projection_mode,
            rotate=rotate,
            scale=scale,
            shear=shear,
            translate=translate,
            units=units,
            visible=visible,
        )

        self.events.add(
            tail_width=Event,
            tail_length=Event,
            head_length=Event,
            display_id=Event,
            display_tail=Event,
            display_graph=Event,
            color_by=Event,
            colormap=Event,
            properties=Event,
            rebuild_tracks=Event,
            rebuild_graph=Event,
        )

        # track manager deals with data slicing, graph building and properties
        self._manager = TrackManager(data)

        self._track_colors: np.ndarray | None = None
        self._colormaps_dict = colormaps_dict or {}  # additional colormaps
        self._color_by = color_by  # default color by ID
        self._colormap = colormap

        # use this to update shaders when the displayed dims change
        self._current_displayed_dims = None

        # track display default limits
        self._max_length = 300
        self._max_width = 20

        # track display properties
        self.tail_width = tail_width
        self.tail_length = tail_length
        self.head_length = head_length
        self.display_id = False
        self.display_tail = True
        self.display_graph = True

        # set the data, features, and graph
        self.data = data
        if properties is not None:
            self.properties = properties
        else:
            self.features = features
        self.graph = graph or {}

        self.color_by = color_by
        self.colormap = colormap

        self.refresh()

        # reset the display before returning
        self._current_displayed_dims = None

    @property
    def _extent_data(self) -> np.ndarray:
        """Extent of layer in data coordinates.

        Returns
        -------
        extent_data : array, shape (2, D)
        """
        if len(self.data) == 0:
            extrema = np.full((2, self.ndim), np.nan)
        else:
            maxs = np.max(self.data, axis=0)
            mins = np.min(self.data, axis=0)
            extrema = np.vstack([mins, maxs])
        return extrema[:, 1:]

    def _get_ndim(self) -> int:
        """Determine number of dimensions of the layer."""
        return self._manager.ndim

    def _get_state(self) -> dict[str, Any]:
        """Get dictionary of layer state.

        Returns
        -------
        state : dict of str to Any
            Dictionary of layer state.
        """
        state = self._get_base_state()
        state.update(
            {
                'data': self.data,
                'properties': self.properties,
                'graph': self.graph,
                'color_by': self.color_by,
                'colormap': self.colormap,
                'colormaps_dict': self.colormaps_dict,
                'tail_width': self.tail_width,
                'tail_length': self.tail_length,
                'head_length': self.head_length,
                'features': self.features,
            }
        )
        return state

    def _set_view_slice(self) -> None:
        """Sets the view given the indices to slice with."""

        # if the displayed dims have changed, update the shader data
        dims_displayed = self._slice_input.displayed
        if dims_displayed != self._current_displayed_dims:
            # store the new dims
            self._current_displayed_dims = dims_displayed
            # fire the events to update the shaders
            self.events.rebuild_tracks()
            self.events.rebuild_graph()

        return

    def _get_value(self, position) -> int | None:
        """Value of the data at a position in data coordinates.

        Use a kd-tree to lookup the ID of the nearest tree.

        Parameters
        ----------
        position : tuple
            Position in data coordinates.

        Returns
        -------
        value : int or None
            Index of track that is at the current coordinate if any.
        """
        val = self._manager.get_value(np.array(position))
        if val is None:
            return None
        return int(val)

    def _update_thumbnail(self) -> None:
        """Update thumbnail with current points and colors."""
        colormapped = np.zeros(self._thumbnail_shape)
        colormapped[..., 3] = 1

        if self._view_data is not None and self.track_colors is not None:
            de = self._extent_data
            min_vals = [de[0, i] for i in self._slice_input.displayed]
            shape = np.ceil(
                [de[1, i] - de[0, i] + 1 for i in self._slice_input.displayed]
            ).astype(int)
            zoom_factor = np.divide(
                self._thumbnail_shape[:2], shape[-2:]
            ).min()
            if len(self._view_data) > self._max_tracks_thumbnail:
                thumbnail_indices = np.random.randint(
                    0, len(self._view_data), self._max_tracks_thumbnail
                )
                points = self._view_data[thumbnail_indices]
            else:
                points = self._view_data
                thumbnail_indices = np.array(range(len(self._view_data)))

            # get the track coords here
            coords = np.floor(
                (points[:, :2] - min_vals[1:] + 0.5) * zoom_factor
            ).astype(int)
            coords = np.clip(
                coords, 0, np.subtract(self._thumbnail_shape[:2], 1)
            )

            # modulate track colors as per colormap/current_time
            assert self.track_times is not None
            assert self.current_time is not None
            colors = self.track_colors[thumbnail_indices]
            times = self.track_times[thumbnail_indices]
            alpha = (self.head_length + self.current_time - times) / (
                self.tail_length + self.head_length
            )
            alpha[times > self.current_time] = 1.0
            colors[:, -1] = np.clip(1.0 - alpha, 0.0, 1.0)
            colormapped[coords[:, 1], coords[:, 0]] = colors

        colormapped[..., 3] *= self.opacity
        colormapped[np.isnan(colormapped)] = 0
        self.thumbnail = colormapped.astype(np.uint8)

    @property
    def _view_data(self):
        """return a view of the data"""
        return self._pad_display_data(self._manager.track_vertices)

    @property
    def _view_graph(self):
        """return a view of the graph"""
        return self._pad_display_data(self._manager.graph_vertices)

    def _pad_display_data(self, vertices):
        """pad display data when moving between 2d and 3d"""
        if vertices is None:
            return None

        data = vertices[:, self._slice_input.displayed]
        # if we're only displaying two dimensions, then pad the display dim
        # with zeros
        if self._slice_input.ndisplay == 2:
            data = np.pad(data, ((0, 0), (0, 1)), 'constant')
            return data[:, (1, 0, 2)]  # y, x, z -> x, y, z

        return data[:, (2, 1, 0)]  # z, y, x -> x, y, z

    @property
    def current_time(self) -> int | None:
        """current time according to the first dimension"""
        # TODO(arl): get the correct index here
        time_step = self._data_slice.point[0]

        if isinstance(time_step, slice):
            # if we are visualizing all time, then just set to the maximum
            # timestamp of the dataset
            return self._manager.max_time

        return time_step

    @property
    def use_fade(self) -> bool:
        """toggle whether we fade the tail of the track, depending on whether
        the time dimension is displayed"""
        return 0 in self._slice_input.not_displayed

    @property
    def data(self) -> np.ndarray:
        """array (N, D+1): Coordinates for N points in D+1 dimensions."""
        return self._manager.data

    @data.setter
    def data(self, data: np.ndarray) -> None:
        """set the data and build the vispy arrays for display"""
        # set the data and build the tracks
        self._manager.data = data
        self._manager.build_tracks()

        # reset the properties and recolor the tracks
        self.features = {}
        self._recolor_tracks()

        # reset the graph
        self._manager.graph = {}
        self._manager.build_graph()

        # fire events to update shaders
        self._update_dims()
        self.events.rebuild_tracks()
        self.events.rebuild_graph()
        self.events.data(value=self.data)
        self._reset_editable()

    @property
    def features(self) -> pd.DataFrame:
        """Dataframe-like features table.

        It is an implementation detail that this is a `pandas.DataFrame`. In the future,
        we will target the currently-in-development Data API dataframe protocol [1].
        This will enable us to use alternate libraries such as xarray or cuDF for
        additional features without breaking existing usage of this.

        If you need to specifically rely on the pandas API, please coerce this to a
        `pandas.DataFrame` using `features_to_pandas_dataframe`.

        References
        ----------
        .. [1]: https://data-apis.org/dataframe-protocol/latest/API.html
        """
        return self._manager.features

    @features.setter
    def features(
        self,
        features: dict[str, np.ndarray] | pd.DataFrame,
    ) -> None:
        self._manager.features = features
        self._check_color_by_in_features()
        self.events.properties()

    @property
    def properties(self) -> dict[str, np.ndarray]:
        """dict {str: np.ndarray (N,)}: Properties for each track."""
        return self._manager.properties

    @properties.setter
    def properties(self, properties: dict[str, np.ndarray]) -> None:
        """set track properties"""
        self.features = properties

    @property
    def properties_to_color_by(self) -> list[str]:
        """track properties that can be used for coloring etc..."""
        return list(self.properties.keys())

    @property
    def graph(self) -> dict[int, list[int]] | None:
        """dict {int: list}: Graph representing associations between tracks."""
        return self._manager.graph

    @graph.setter
    def graph(self, graph: dict[int, int | list[int]]) -> None:
        """Set the track graph."""
        # Ignored type, because mypy can't handle different signatures
        # on getters and setters; see https://github.com/python/mypy/issues/3004
        self._manager.graph = graph  # type: ignore[assignment]
        self._manager.build_graph()
        self.events.rebuild_graph()

    @property
    def tail_width(self) -> float:
        """float: Width for all vectors in pixels."""
        return self._tail_width

    @tail_width.setter
    def tail_width(self, tail_width: float) -> None:
        self._tail_width: float = np.clip(tail_width, 0.5, self._max_width)
        self.events.tail_width()

    @property
    def tail_length(self) -> int:
        """float: Width for all vectors in pixels."""
        return self._tail_length

    @tail_length.setter
    def tail_length(self, tail_length: int) -> None:
        if tail_length > self._max_length:
            self._max_length = tail_length
        self._tail_length: int = tail_length
        self.events.tail_length()

    @property
    def head_length(self) -> int:
        return self._head_length

    @head_length.setter
    def head_length(self, head_length: int) -> None:
        if head_length > self._max_length:
            self._max_length = head_length
        self._head_length: int = head_length
        self.events.head_length()

    @property
    def display_id(self) -> bool:
        """display the track id"""
        return self._display_id

    @display_id.setter
    def display_id(self, value: bool) -> None:
        self._display_id = value
        self.events.display_id()
        # TODO: this refresh is only here to trigger setting the id text...
        #       a bit overkill? But maybe for a future PR.
        self.refresh(extent=False, thumbnail=False)

    @property
    def display_tail(self) -> bool:
        """display the track tail"""
        return self._display_tail

    @display_tail.setter
    def display_tail(self, value: bool) -> None:
        self._display_tail = value
        self.events.display_tail()

    @property
    def display_graph(self) -> bool:
        """display the graph edges"""
        return self._display_graph

    @display_graph.setter
    def display_graph(self, value: bool) -> None:
        self._display_graph = value
        self.events.display_graph()

    @property
    def color_by(self) -> str:
        return self._color_by

    @color_by.setter
    def color_by(self, color_by: str) -> None:
        """set the property to color vertices by"""
        if color_by not in self.properties_to_color_by:
            raise ValueError(
                trans._(
                    '{color_by} is not a valid property key',
                    deferred=True,
                    color_by=color_by,
                )
            )
        self._color_by = color_by
        self._recolor_tracks()
        self.events.color_by()

    @property
    def colormap(self) -> str:
        return self._colormap

    @colormap.setter
    def colormap(self, colormap: str) -> None:
        """set the default colormap"""
        if colormap not in AVAILABLE_COLORMAPS:
            raise ValueError(
                trans._(
                    'Colormap {colormap} not available',
                    deferred=True,
                    colormap=colormap,
                )
            )
        self._colormap = colormap
        self._recolor_tracks()
        self.events.colormap()

    @property
    def colormaps_dict(self) -> dict[str, Colormap]:
        return self._colormaps_dict

    @colormaps_dict.setter
    def colormaps_dict(self, colormaps_dict: dict[str, Colormap]) -> None:
        # validate the dictionary entries?
        self._colormaps_dict = colormaps_dict

    def _recolor_tracks(self) -> None:
        """recolor the tracks"""

        # this catch prevents a problem coloring the tracks if the data is
        # updated before the properties are. properties should always contain
        # a track_id key
        if self.color_by not in self.properties_to_color_by:
            self._color_by = 'track_id'
            self.events.color_by()

        # if we change the coloring, rebuild the vertex colors array
        vertex_properties = self._manager.vertex_properties(self.color_by)

        def _norm(p):
            return (p - np.min(p)) / np.max([1e-10, np.ptp(p)])

        if self.color_by in self.colormaps_dict:
            colormap = self.colormaps_dict[self.color_by]
        else:
            # if we don't have a colormap, get one and scale the properties
            colormap = AVAILABLE_COLORMAPS[self.colormap]
            vertex_properties = _norm(vertex_properties)

        # actually set the vertex colors
        self._track_colors = colormap.map(vertex_properties)

    @property
    def track_connex(self) -> np.ndarray | None:
        """vertex connections for drawing track lines"""
        return self._manager.track_connex

    @property
    def track_colors(self) -> np.ndarray | None:
        """Return current vertex colors."""
        return self._track_colors

    @track_colors.setter
    def track_colors(self, colors: np.ndarray) -> None:
        """Set vertex colors directly and emit corresponding event.

        Parameters
        ----------
        colors : np.ndarray, shape (N, 4)
            The desired color array, one color per data point. Note that the
            data are internally lexicographically sorted first by track ID,
            then by timepoint, so the color ordering must match this order,
            which may not be the input order. Use `Tracks.data` to find the
            current sorted data.
        """
        self._track_colors = colors
        self.events.color_by()

    @property
    def graph_connex(self) -> np.ndarray | None:
        """vertex connections for drawing the graph"""
        return self._manager.graph_connex

    @property
    def track_times(self) -> np.ndarray | None:
        """time points associated with each track vertex"""
        return self._manager.track_times

    @property
    def graph_times(self) -> np.ndarray | None:
        """time points associated with each graph vertex"""
        return self._manager.graph_times

    @property
    def track_labels(self) -> tuple:
        """return track labels at the current time"""
        assert self.current_time is not None
        labels, positions = self._manager.track_labels(self.current_time)

        # if there are no labels, return empty for vispy
        if not labels:
            return None, (None, None)

        padded_positions = self._pad_display_data(positions)
        return labels, padded_positions

    def _check_color_by_in_features(self) -> None:
        if self._color_by not in self.features.columns:
            warn(
                (
                    trans._(
                        'Previous color_by key {key!r} not present in features. Falling back to track_id',
                        deferred=True,
                        key=self._color_by,
                    )
                ),
                UserWarning,
            )
            self._color_by = 'track_id'
            self.events.color_by()



import numpy as np

from napari._vispy.overlays.base import ViewerOverlayMixin, VispySceneOverlay
from napari._vispy.visuals.axes import Axes
from napari.utils.theme import get_theme


class VispyAxesOverlay(ViewerOverlayMixin, VispySceneOverlay):
    """Axes indicating world coordinate origin and orientation."""

    def __init__(self, *, viewer, overlay, parent=None) -> None:
        self._scale = 1

        # Target axes length in canvas pixels
        self._target_length = 80

        super().__init__(
            node=Axes(), viewer=viewer, overlay=overlay, parent=parent
        )
        self.overlay.events.colored.connect(self._on_data_change)
        self.overlay.events.dashed.connect(self._on_data_change)
        self.overlay.events.labels.connect(self._on_labels_visible_change)
        self.overlay.events.arrows.connect(self._on_data_change)

        self.viewer.events.theme.connect(self._on_data_change)
        self.viewer.camera.events.zoom.connect(self._on_zoom_change)
        self.viewer.dims.events.order.connect(self._on_data_change)
        self.viewer.dims.events.range.connect(self._on_data_change)
        self.viewer.dims.events.ndisplay.connect(self._on_data_change)
        self.viewer.dims.events.axis_labels.connect(
            self._on_labels_text_change
        )

        self.reset()

    def _on_data_change(self):
        # Determine which axes are displayed
        axes = self.viewer.dims.displayed[::-1]

        # Counting backwards from total number of dimensions
        # determine axes positions. This is done as by default
        # the last NumPy axis corresponds to the first Vispy axis
        reversed_axes = [self.viewer.dims.ndim - 1 - a for a in axes]

        self.node.set_data(
            axes=axes,
            reversed_axes=reversed_axes,
            colored=self.overlay.colored,
            bg_color=get_theme(self.viewer.theme).canvas,
            dashed=self.overlay.dashed,
            arrows=self.overlay.arrows,
        )

        self._on_labels_text_change()

    def _on_labels_visible_change(self):
        self.node.text.visible = self.overlay.labels

    def _on_labels_text_change(self):
        axes = self.viewer.dims.displayed[::-1]
        axis_labels = [self.viewer.dims.axis_labels[a] for a in axes]
        self.node.text.text = axis_labels

    def _on_zoom_change(self):
        scale = 1 / self.viewer.camera.zoom

        # If scale has not changed, do not redraw
        if abs(np.log10(self._scale) - np.log10(scale)) < 1e-4:
            return
        self._scale = scale
        scale = self._target_length * self._scale
        # Update axes scale
        self.node.transform.reset()
        self.node.transform.scale([scale, scale, scale, 1])

    def reset(self):
        super().reset()
        self._on_data_change()
        self._on_labels_visible_change()
        self._on_zoom_change()



import inspect
import warnings
from collections.abc import Callable, Sequence
from functools import partial, wraps
from types import FunctionType, GeneratorType
from typing import (
    TypeVar,
)

from superqt.utils import _qthreading

from napari.utils.progress import progress
from napari.utils.translations import trans

__all__ = [
    'FunctionWorker',
    'GeneratorWorker',
    'create_worker',
    'register_threadworker_processors',
    'thread_worker',
]

wait_for_workers_to_quit = _qthreading.WorkerBase.await_workers


class _NotifyingMixin:
    def __init__(self: _qthreading.WorkerBase, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)  # type: ignore
        self.errored.connect(self._relay_error)
        self.warned.connect(self._relay_warning)

    def _relay_error(self, exc: Exception):
        from napari.utils.notifications import notification_manager

        notification_manager.receive_error(type(exc), exc, exc.__traceback__)

    def _relay_warning(self, show_warn_args: tuple):
        from napari.utils.notifications import notification_manager

        notification_manager.receive_warning(*show_warn_args)


_Y = TypeVar('_Y')
_S = TypeVar('_S')
_R = TypeVar('_R')


class FunctionWorker(_qthreading.FunctionWorker[_R], _NotifyingMixin): ...


class GeneratorWorker(
    _qthreading.GeneratorWorker[_Y, _S, _R], _NotifyingMixin
): ...


# these are re-implemented from superqt just to provide progress


def create_worker(
    func: FunctionType | GeneratorType,
    *args,
    _start_thread: bool | None = None,
    _connect: dict[str, Callable | Sequence[Callable]] | None = None,
    _progress: bool | dict[str, int | bool | str] | None = None,
    _worker_class: type[GeneratorWorker] | type[FunctionWorker] | None = None,
    _ignore_errors: bool = False,
    **kwargs,
) -> FunctionWorker | GeneratorWorker:
    """Convenience function to start a function in another thread.

    By default, uses :class:`Worker`, but a custom ``WorkerBase`` subclass may
    be provided.  If so, it must be a subclass of :class:`Worker`, which
    defines a standard set of signals and a run method.

    Parameters
    ----------
    func : Callable
        The function to call in another thread.
    _start_thread : bool, optional
        Whether to immediaetly start the thread.  If False, the returned worker
        must be manually started with ``worker.start()``. by default it will be
        ``False`` if the ``_connect`` argument is ``None``, otherwise ``True``.
    _connect : Dict[str, Union[Callable, Sequence]], optional
        A mapping of ``"signal_name"`` -> ``callable`` or list of ``callable``:
        callback functions to connect to the various signals offered by the
        worker class. by default None
    _progress : Union[bool, Dict[str, Union[int, bool, str]]], optional
        Can be True, to provide indeterminate progress bar, or dictionary.
        If dict, requires mapping of 'total' to number of expected yields.
        If total is not provided, progress bar will be indeterminate. Will connect
        progress bar update to yields and display this progress in the viewer.
        Can also take a mapping of 'desc' to the progress bar description.
        Progress bar will become indeterminate when number of yields exceeds 'total'.
        By default None.
    _worker_class : Type[WorkerBase], optional
        The :class`WorkerBase` to instantiate, by default
        :class:`FunctionWorker` will be used if ``func`` is a regular function,
        and :class:`GeneratorWorker` will be used if it is a generator.
    _ignore_errors : bool, optional
        If ``False`` (the default), errors raised in the other thread will be
        reraised in the main thread (makes debugging significantly easier).
    *args
        will be passed to ``func``
    **kwargs
        will be passed to ``func``

    Returns
    -------
    worker : WorkerBase
        An instantiated worker.  If ``_start_thread`` was ``False``, the worker
        will have a `.start()` method that can be used to start the thread.

    Raises
    ------
    TypeError
        If a worker_class is provided that is not a subclass of WorkerBase.
    TypeError
        If _connect is provided and is not a dict of ``{str: callable}``
    TypeError
        If _progress is provided and function is not a generator

    Examples
    --------
    .. code-block:: python

        def long_function(duration):
            import time
            time.sleep(duration)

        worker = create_worker(long_function, 10)

    """
    # provide our own classes with the notification mixins
    if not _worker_class:
        if inspect.isgeneratorfunction(func):
            _worker_class = GeneratorWorker
        else:
            _worker_class = FunctionWorker

    worker = _qthreading.create_worker(
        func,
        *args,
        _start_thread=False,
        _connect=_connect,
        _worker_class=_worker_class,
        _ignore_errors=_ignore_errors,
        **kwargs,
    )

    # either True or a non-empty dictionary
    if _progress:
        if isinstance(_progress, bool):
            _progress = {}

        desc = _progress.get('desc', None)
        total = int(_progress.get('total', 0))
        if isinstance(worker, FunctionWorker) and total != 0:
            warnings.warn(
                trans._(
                    '_progress total != 0 but worker is FunctionWorker and will not yield. Returning indeterminate progress bar...',
                    deferred=True,
                ),
                RuntimeWarning,
            )
            total = 0

        with progress._all_instances.events.changed.blocker():
            pbar = progress(total=total, desc=desc)

        worker.started.connect(
            partial(
                lambda prog: progress._all_instances.events.changed(
                    added={prog}, removed={}
                ),
                pbar,
            )
        )
        worker.finished.connect(pbar.close)
        if total != 0 and isinstance(worker, GeneratorWorker):
            worker.yielded.connect(pbar.increment_with_overflow)

        worker.pbar = pbar

    if _start_thread is None:
        _start_thread = _connect is not None

    if _start_thread:
        worker.start()
    return worker


def thread_worker(
    function: Callable | None = None,
    start_thread: bool | None = None,
    connect: dict[str, Callable | Sequence[Callable]] | None = None,
    progress: bool | dict[str, int | bool | str] | None = None,
    worker_class: type[FunctionWorker] | type[GeneratorWorker] | None = None,
    ignore_errors: bool = False,
):
    """Decorator that runs a function in a separate thread when called.

    When called, the decorated function returns a :class:`WorkerBase`.  See
    :func:`create_worker` for additional keyword arguments that can be used
    when calling the function.

    The returned worker will have these signals:

    - *started*: emitted when the work is started
    - *finished*: emitted when the work is finished
    - *returned*: emitted with return value
    - *errored*: emitted with error object on Exception

    It will also have a ``worker.start()`` method that can be used to start
    execution of the function in another thread. (useful if you need to connect
    callbacks to signals prior to execution)

    If the decorated function is a generator, the returned worker will also
    provide these signals:

    - *yielded*: emitted with yielded values
    - *paused*: emitted when a running job has successfully paused
    - *resumed*: emitted when a paused job has successfully resumed
    - *aborted*: emitted when a running job is successfully aborted

    And these methods:

    - *quit*: ask the thread to quit
    - *toggle_paused*: toggle the running state of the thread.
    - *send*: send a value into the generator.  (This requires that your
        decorator function uses the ``value = yield`` syntax)

    Parameters
    ----------
    function : callable
        Function to call in another thread.  For communication between threads
        may be a generator function.
    start_thread : bool, optional
        Whether to immediaetly start the thread.  If False, the returned worker
        must be manually started with ``worker.start()``. by default it will be
        ``False`` if the ``_connect`` argument is ``None``, otherwise ``True``.
    connect : Dict[str, Union[Callable, Sequence]], optional
        A mapping of ``"signal_name"`` -> ``callable`` or list of ``callable``:
        callback functions to connect to the various signals offered by the
        worker class. by default None
    progress : Union[bool, Dict[str, Union[int, bool, str]]], optional
        Can be True, to provide indeterminate progress bar, or dictionary.
        If dict, requires mapping of 'total' to number of expected yields.
        If total is not provided, progress bar will be indeterminate. Will connect
        progress bar update to yields and display this progress in the viewer.
        Can also take a mapping of 'desc' to the progress bar description.
        Progress bar will become indeterminate when number of yields exceeds 'total'.
        By default None. Must be used in conjunction with a generator function.
    worker_class : Type[WorkerBase], optional
        The :class`WorkerBase` to instantiate, by default
        :class:`FunctionWorker` will be used if ``func`` is a regular function,
        and :class:`GeneratorWorker` will be used if it is a generator.
    ignore_errors : bool, optional
        If ``False`` (the default), errors raised in the other thread will be
        reraised in the main thread (makes debugging significantly easier).

    Returns
    -------
    callable
        function that creates a worker, puts it in a new thread and returns
        the worker instance.

    Examples
    --------
    .. code-block:: python

        @thread_worker
        def long_function(start, end):
            # do work, periodically yielding
            i = start
            while i <= end:
                time.sleep(0.1)
                yield i

            # do teardown
            return 'anything'

        # call the function to start running in another thread.
        worker = long_function()
        # connect signals here if desired... or they may be added using the
        # `connect` argument in the `@thread_worker` decorator... in which
        # case the worker will start immediately when long_function() is called
        worker.start()
    """

    def _inner(func):
        @wraps(func)
        def worker_function(*args, **kwargs):
            # decorator kwargs can be overridden at call time by using the
            # underscore-prefixed version of the kwarg.
            kwargs['_start_thread'] = kwargs.get('_start_thread', start_thread)
            kwargs['_connect'] = kwargs.get('_connect', connect)
            kwargs['_progress'] = kwargs.get('_progress', progress)
            kwargs['_worker_class'] = kwargs.get('_worker_class', worker_class)
            kwargs['_ignore_errors'] = kwargs.get(
                '_ignore_errors', ignore_errors
            )
            return create_worker(
                func,
                *args,
                **kwargs,
            )

        return worker_function

    return _inner if function is None else _inner(function)


_new_worker_qthread = _qthreading.new_worker_qthread


def _add_worker_data(worker: FunctionWorker, return_type, source=None):
    from napari._qt._qapp_model.injection._qprocessors import (
        _add_layer_data_to_viewer,
    )

    cb = _add_layer_data_to_viewer
    worker.signals.returned.connect(
        partial(cb, return_type=return_type, source=source)
    )


def _add_worker_data_from_tuple(
    worker: FunctionWorker, return_type, source=None
):
    from napari._qt._qapp_model.injection._qprocessors import (
        _add_layer_data_tuples_to_viewer,
    )

    cb = _add_layer_data_tuples_to_viewer
    worker.signals.returned.connect(
        partial(cb, return_type=return_type, source=source)
    )


def register_threadworker_processors():
    from functools import partial

    import magicgui

    from napari import layers, types
    from napari._app_model import get_app_model
    from napari.types import LayerDataTuple
    from napari.utils import _magicgui as _mgui

    app = get_app_model()

    for _type in (LayerDataTuple, list[LayerDataTuple]):
        t = FunctionWorker[_type]
        magicgui.register_type(t, return_callback=_mgui.add_worker_data)
        app.injection_store.register(
            processors={t: _add_worker_data_from_tuple}
        )
    for layer_name in layers.NAMES:
        _type = getattr(types, f'{layer_name.title()}Data')
        t = FunctionWorker[_type]
        magicgui.register_type(
            t,
            return_callback=partial(_mgui.add_worker_data, _from_tuple=False),
        )
        app.injection_store.register(processors={t: _add_worker_data})



"""For testing the Help menu"""

import sys
from unittest import mock

import pytest
import requests

from napari._app_model import get_app_model
from napari._qt._qapp_model.qactions._help import HELP_URLS


@pytest.mark.parametrize('url', HELP_URLS.keys())
def test_help_urls(url):
    if url == 'release_notes':
        pytest.skip('No release notes for dev version')

    r = requests.head(HELP_URLS[url])
    r.raise_for_status()


@pytest.mark.parametrize(
    'action_id',
    [
        'napari.window.help.info',
        'napari.window.help.about_macos',
    ]
    if sys.platform == 'darwin'
    else ['napari.window.help.info'],
)
def test_about_action(make_napari_viewer, action_id):
    app = get_app_model()
    viewer = make_napari_viewer()

    with mock.patch(
        'napari._qt.dialogs.qt_about.QtAbout.showAbout'
    ) as mock_about:
        app.commands.execute_command(action_id)
    mock_about.assert_called_once_with(viewer.window._qt_window)



import napari.utils.notifications
from napari._qt.qt_event_loop import run
from napari.plugins.io import save_layers
from napari.view_layers import (
    view_image,
    view_labels,
    view_path,
    view_points,
    view_shapes,
    view_surface,
    view_tracks,
    view_vectors,
)
from napari.viewer import Viewer, current_viewer

__version__: str

notification_manager: napari.utils.notifications.NotificationManager

__all__ = (
    'Viewer',
    '__version__',
    'current_viewer',
    'notification_manager',
    'run',
    'save_layers',
    'view_image',
    'view_labels',
    'view_path',
    'view_points',
    'view_shapes',
    'view_surface',
    'view_tracks',
    'view_vectors',
)



import numpy as np
import numpy.testing as npt
import pytest

from napari.utils.colormaps.colormap_utils import (
    CoercedContrastLimits,
    _coerce_contrast_limits,
    label_colormap,
)

FIRST_COLORS = [
    [0.47058824, 0.14509805, 0.02352941, 1.0],
    [0.35686275, 0.8352941, 0.972549, 1.0],
    [0.57254905, 0.5372549, 0.9098039, 1.0],
    [0.42352942, 0.00784314, 0.75686276, 1.0],
    [0.2784314, 0.22745098, 0.62352943, 1.0],
    [0.67058825, 0.9254902, 0.5411765, 1.0],
    [0.56078434, 0.6784314, 0.69803923, 1.0],
    [0.5254902, 0.5647059, 0.6039216, 1.0],
    [0.99607843, 0.96862745, 0.10980392, 1.0],
    [0.96862745, 0.26666668, 0.23137255, 1.0],
]


@pytest.mark.parametrize(
    ('index', 'expected'), enumerate(FIRST_COLORS, start=1)
)
def test_label_colormap(index, expected):
    """Test the label colormap.

    Make sure that the default label colormap colors are identical
    to past versions, for UX consistency.
    """
    np.testing.assert_almost_equal(label_colormap(49).map(index), expected)


def test_label_colormap_exception():
    with pytest.raises(ValueError, match='num_colors must be >= 1'):
        label_colormap(0)

    with pytest.raises(ValueError, match='num_colors must be >= 1'):
        label_colormap(-1)

    with pytest.raises(
        ValueError, match=r'.*Only up to 2\*\*16=65535 colors are supported'
    ):
        label_colormap(2**16 + 1)


def test_coerce_contrast_limits_with_valid_input():
    contrast_limits = (0.0, 1.0)
    result = _coerce_contrast_limits(contrast_limits)
    assert isinstance(result, CoercedContrastLimits)
    assert np.allclose(result.contrast_limits, contrast_limits)
    assert result.offset == 0
    assert np.isclose(result.scale, 1.0)
    npt.assert_allclose(
        result.contrast_limits, result.coerce_data(np.array(contrast_limits))
    )


def test_coerce_contrast_limits_with_large_values():
    contrast_limits = (0, float(np.finfo(np.float32).max) * 100)
    result = _coerce_contrast_limits(contrast_limits)
    assert isinstance(result, CoercedContrastLimits)
    assert np.isclose(result.contrast_limits[0], np.finfo(np.float32).min / 8)
    assert np.isclose(result.contrast_limits[1], np.finfo(np.float32).max / 8)
    assert result.offset < 0
    assert result.scale < 1.0
    npt.assert_allclose(
        result.contrast_limits, result.coerce_data(np.array(contrast_limits))
    )


def test_coerce_contrast_limits_with_large_values_symmetric():
    above_float32_max = float(np.finfo(np.float32).max) * 100
    contrast_limits = (-above_float32_max, above_float32_max)
    result = _coerce_contrast_limits(contrast_limits)
    assert isinstance(result, CoercedContrastLimits)
    assert np.isclose(result.contrast_limits[0], np.finfo(np.float32).min / 8)
    assert np.isclose(result.contrast_limits[1], np.finfo(np.float32).max / 8)
    assert result.offset == 0
    assert result.scale < 1.0
    npt.assert_allclose(
        result.contrast_limits, result.coerce_data(np.array(contrast_limits))
    )


def test_coerce_contrast_limits_with_large_values_above_limit():
    f32_max = float(np.finfo(np.float32).max)
    contrast_limits = (f32_max * 10, f32_max * 100)
    result = _coerce_contrast_limits(contrast_limits)
    assert isinstance(result, CoercedContrastLimits)
    assert np.isclose(result.contrast_limits[0], np.finfo(np.float32).min / 8)
    assert np.isclose(result.contrast_limits[1], np.finfo(np.float32).max / 8)
    assert result.offset < 0
    assert result.scale < 1.0
    npt.assert_allclose(
        result.contrast_limits, result.coerce_data(np.array(contrast_limits))
    )


def test_coerce_contrast_limits_small_values():
    contrast_limits = (1e-45, 9e-45)
    result = _coerce_contrast_limits(contrast_limits)
    assert isinstance(result, CoercedContrastLimits)
    assert np.isclose(result.contrast_limits[0], 0)
    assert np.isclose(result.contrast_limits[1], 1000)
    assert result.offset < 0
    assert result.scale > 1
    npt.assert_allclose(
        result.contrast_limits, result.coerce_data(np.array(contrast_limits))
    )



import numpy as np
import pytest

from napari.utils.colormaps._tests.colors_data import (
    invalid_colors,
    single_color_options,
    single_colors_as_array,
    two_color_options,
    two_colors_as_array,
    warning_colors,
)
from napari.utils.colormaps.standardize_color import transform_color


@pytest.mark.parametrize(
    ('colors', 'true_colors'),
    zip(single_color_options, single_colors_as_array, strict=False),
)
def test_oned_points(colors, true_colors):
    np.testing.assert_array_equal(true_colors, transform_color(colors))


def test_warns_but_parses():
    """Test collection of colors that raise a warning but do not return
    a default white color array.
    """
    colors = ['', (43, 3, 3, 3), np.array([[3, 3, 3, 3], [0, 0, 0, 1]])]

    true_colors = [
        np.zeros((1, 4), dtype=np.float32),
        np.array([[1, 3 / 43, 3 / 43, 3 / 43]], dtype=np.float32),
        np.array(
            [[1.0, 1.0, 1.0, 1.0], [0.0, 0.0, 0.0, 1.0]], dtype=np.float32
        ),
    ]
    with pytest.warns(UserWarning):
        for true, color in zip(true_colors, colors, strict=False):
            np.testing.assert_array_equal(true, transform_color(color))


@pytest.mark.parametrize(
    ('colors', 'true_colors'),
    zip(two_color_options, two_colors_as_array, strict=False),
)
def test_twod_points(colors, true_colors):
    np.testing.assert_array_equal(true_colors, transform_color(colors))


@pytest.mark.parametrize('color', invalid_colors)
def test_invalid_colors(color):
    with pytest.raises((ValueError, AttributeError, KeyError)):
        transform_color(color)


@pytest.mark.parametrize('colors', warning_colors)
def test_warning_colors(colors):
    with pytest.warns(UserWarning):
        np.testing.assert_array_equal(
            np.ones((max(len(colors), 1), 4), dtype=np.float32),
            transform_color(colors),
        )



import pytest

from napari.utils.events.migrations import deprecation_warning_event


def test_deprecation_warning_event() -> None:
    event = deprecation_warning_event(
        'obj.events', 'old', 'new', '0.1.0', '0.0.0'
    )

    class Counter:
        def __init__(self) -> None:
            self.count = 0

        def add(self, event) -> None:
            self.count += event.value

    counter = Counter()
    msg = 'obj.events.old is deprecated since 0.0.0 and will be removed in 0.1.0. Please use obj.events.new'

    with pytest.warns(FutureWarning, match=msg):
        event.connect(counter.add)
        event(value=1)

    assert counter.count == 1



import os
import sys
from importlib.metadata import PackageNotFoundError, distribution
from unittest.mock import patch

import pytest

from napari.settings import NapariSettings, _migrations
from napari.utils.triangulation_backend import TriangulationBackend


@pytest.fixture
def test_migrator(monkeypatch):
    # this fixture makes sure we're not using _migrations.MIGRATORS for tests
    # but rather only using migrators that get declared IN the test
    _TEST_MIGRATORS = []
    with monkeypatch.context() as m:
        m.setattr(_migrations, '_MIGRATORS', _TEST_MIGRATORS)
        yield _migrations.migrator


def test_no_migrations_available(test_migrator):
    # no migrators exist... nothing should happen
    settings = NapariSettings(schema_version='0.1.0')
    assert settings.schema_version == '0.1.0'


def test_backwards_migrator(test_migrator):
    # we shouldn't be able to downgrade the schema version
    # if that is needed later, we can create a new decorator,
    # or change this test
    with pytest.raises(AssertionError):

        @test_migrator('0.2.0', '0.1.0')
        def _(model): ...


def test_migration_works(test_migrator):
    # test that a basic migrator works to change the version
    # and mutate the model

    @test_migrator('0.1.0', '0.2.0')
    def _(model: NapariSettings):
        model.appearance.theme = 'light'

    settings = NapariSettings(schema_version='0.1.0')
    assert settings.schema_version == '0.2.0'
    assert settings.appearance.theme == 'light'


def test_migration_saves(test_migrator):
    @test_migrator('0.1.0', '0.2.0')
    def _(model: NapariSettings): ...

    with patch.object(NapariSettings, 'save') as mock:
        mock.assert_not_called()
        settings = NapariSettings(config_path='junk', schema_version='0.1.0')
        assert settings.schema_version == '0.2.0'
        mock.assert_called()


def test_failed_migration_leaves_version(test_migrator):
    # if an error occurs IN the migrator, the version should stay
    # where it was before the migration, and any changes reverted.
    @test_migrator('0.1.0', '0.2.0')
    def _(model: NapariSettings):
        model.appearance.theme = 'light'
        assert model.appearance.theme == 'light'
        raise ValueError('broken migration')

    with pytest.warns(UserWarning) as e:
        settings = NapariSettings(schema_version='0.1.0')
    assert settings.schema_version == '0.1.0'
    # test migration was atomic, and reverted the theme change
    assert settings.appearance.theme == 'dark'
    # test that the user was warned
    assert 'Failed to migrate settings from v0.1.0 to v0.2.0' in str(e[0])


@pytest.mark.skipif(
    bool(os.environ.get('MIN_REQ')), reason='not relevant for MIN_REQ'
)
def test_030_to_040_migration():
    # Prior to v0.4.0, npe2 plugins were automatically "disabled"
    # 0.3.0 -> 0.4.0 should remove any installed npe2 plugins from the
    # set of disabled plugins (see migrator for details)
    try:
        d = distribution('napari-svg')
        assert 'napari.manifest' in {ep.group for ep in d.entry_points}
    except PackageNotFoundError:
        pytest.fail(
            'napari-svg not present as an npe2 plugin. '
            'This test needs updating'
        )

    settings = NapariSettings(
        schema_version='0.3.0',
        plugins={'disabled_plugins': {'napari-svg', 'napari'}},
    )
    assert 'napari-svg' not in settings.plugins.disabled_plugins
    assert 'napari' not in settings.plugins.disabled_plugins


@pytest.mark.skipif(
    bool(os.environ.get('MIN_REQ')), reason='not relevant for MIN_REQ'
)
def test_040_to_050_migration():
    # Prior to 0.5.0 existing preferences may have reader extensions
    # preferences saved without a leading *.
    # fnmatch would fail on these so we coerce them to include a *
    # e.g. '.csv' becomes '*.csv'
    settings = NapariSettings(
        schema_version='0.4.0',
        plugins={'extension2reader': {'.tif': 'napari'}},
    )
    assert '.tif' not in settings.plugins.extension2reader
    assert '*.tif' in settings.plugins.extension2reader


@pytest.mark.skipif(sys.platform != 'darwin', reason='tests migration on macs')
def test_050_to_060_migration_mac():
    """Check that Ctrl and Meta are swapped on macOS when migrating."""
    settings050 = NapariSettings(
        schema_version='0.5.0',
        shortcuts={
            'shortcuts': {
                'napari:focus_axes_up': ['Alt-Up'],
                'napari:roll_axes': ['Control-E'],
                'napari:transpose_axes': ['Control-Meta-T'],
                'napari:paste_shape': ['V', 'Meta-T'],
            }
        },
    )
    settings060 = NapariSettings(
        schema_version='0.6.0',
        shortcuts={
            'shortcuts': {
                'napari:focus_axes_up': ['Alt-Up'],
                'napari:roll_axes': ['Meta-E'],
                'napari:transpose_axes': ['Ctrl-Meta-T'],
                'napari:paste_shape': ['V', 'Ctrl-T'],
            }
        },
    )
    assert settings050 == settings060


@pytest.mark.skipif(
    sys.platform == 'darwin', reason='migration should not be no-op on macs'
)
def test_050_to_060_migration_linux_win():
    """Check that shortcuts are unchanged on non-macOS when migrating."""
    shortcuts_dict = {
        'napari:focus_axes_up': ['Alt-Up'],
        'napari:roll_axes': ['Control-E'],
        'napari:transpose_axes': ['Control-Meta-T'],
        'napari:paste_shape': ['V', 'Meta-T'],
    }
    settings050 = NapariSettings(
        schema_version='0.5.0', shortcuts={'shortcuts': shortcuts_dict}
    )
    settings060 = NapariSettings(
        schema_version='0.6.0', shortcuts={'shortcuts': shortcuts_dict}
    )
    assert settings050 == settings060


def test_060_to_070_migration():
    settings_060_plain = NapariSettings(schema_version='0.6.0')
    settings_060_compiled = NapariSettings(
        schema_version='0.6.0', experimental={'compiled_triangulation': True}
    )
    settings_070_plain_expected = NapariSettings(
        schema_version='0.7.0',
        experimental={
            # must match default setting
            'triangulation_backend': TriangulationBackend.fastest_available
        },
    )
    settings_070_compiled_expected = NapariSettings(
        schema_version='0.7.0',
        experimental={
            'triangulation_backend': TriangulationBackend.partsegcore
        },
    )
    assert settings_060_plain == settings_070_plain_expected
    assert settings_060_compiled == settings_070_compiled_expected



import pytest

from napari._pydantic_compat import ValidationError
from napari.layers import Points
from napari.layers._source import Source, current_source, layer_source


def test_layer_source():
    """Test basic layer source assignment mechanism"""
    with layer_source(path='some_path', reader_plugin='napari'):
        points = Points()

    assert points.source == Source(path='some_path', reader_plugin='napari')


def test_cant_overwrite_source():
    """Test that we can't overwrite the source of a layer."""
    with layer_source(path='some_path', reader_plugin='napari'):
        points = Points()
    assert points.source == Source(path='some_path', reader_plugin='napari')
    with pytest.raises(ValueError, match='Tried to set source on layer'):
        points._set_source(
            Source(path='other_path', reader_plugin='other_plugin')
        )


def test_source_context():
    """Test nested contexts, overrides, and resets."""

    assert current_source() == Source()
    # everything created within this context will have this sample source
    with layer_source(sample=('samp', 'name')):
        assert current_source() == Source(sample=('samp', 'name'))
        # nested contexts override previous ones
        with layer_source(path='a', reader_plugin='plug'):
            assert current_source() == Source(
                path='a', reader_plugin='plug', sample=('samp', 'name')
            )
            # note the new path now...
            with layer_source(path='b'):
                assert current_source() == Source(
                    path='b', reader_plugin='plug', sample=('samp', 'name')
                )
                # as we exit the contexts, they should undo their assignments
            assert current_source() == Source(
                path='a', reader_plugin='plug', sample=('samp', 'name')
            )
        assert current_source() == Source(sample=('samp', 'name'))

        point = Points()
        with layer_source(parent=point):
            assert current_source() == Source(
                sample=('samp', 'name'), parent=point
            )
    assert current_source() == Source()


def test_source_assert_parent():
    assert current_source() == Source()
    with pytest.raises(ValidationError):
        with layer_source(parent=''):
            current_source()
    assert current_source() == Source()
